//
//  AppDelegate.m
//  C64SceneBrowser
//
//  Created by Andreas Varga on 16.03.12.
//  Copyright (c) 2012 __MyCompanyName__. All rights reserved.
//

#import "AppDelegate.h"
#import "CSDatabaseCache.h"
#import "CSReleaseItem.h"
#import "CSScenerItem.h"
#import "ReleaseViewController.h"
#import "ScenerViewController.h"
#import "EmulationViewController.h"
#import "CreditsViewController.h"
#import "MasterViewController.h"
#import "DetailViewController.h"
#import "UIView+MTDebug.h"

@implementation AppDelegate

@synthesize window = _window;
@synthesize masterPopoverController = _masterPopoverController;
@synthesize masterPopoverBarButtonItem = _masterPopoverBarButtonItem;
@synthesize hasSplitView = _hasSplitView;
@synthesize isInBackground = _isInBackground;


// ----------------------------------------------------------------------------
- (BOOL) application:(UIApplication *)application didFinishLaunchingWithOptions:(NSDictionary *)launchOptions
// ----------------------------------------------------------------------------
{
    //[UIView toggleViewMainThreadChecking];
    
    // Override point for customization after application launch.
    UIViewController* rootViewController = self.window.rootViewController;
    
    if ([rootViewController class] == [UISplitViewController class])
    {
        _hasSplitView = YES;
        UISplitViewController* splitViewController = (UISplitViewController *)rootViewController;

        // Get right view of split view
        UINavigationController* navigationController = [splitViewController.viewControllers lastObject];
        splitViewController.delegate = (id)navigationController.topViewController;
    }
    else
        _hasSplitView = NO;
    
    [[UIApplication sharedApplication] setStatusBarStyle:UIStatusBarStyleLightContent];
    
    //_initialLoadUrl = [launchOptions objectForKey:UIApplicationLaunchOptionsURLKey];
    //_initialLoadUrl = [NSURL URLWithString:@"file:///private/var/mobile/Containers/Data/Application/8242A9D9-DD07-4B58-85EF-D83B42F8E47E/Documents/Inbox/old_farts-74.prg"];
    //_initialLoadUrl = [NSURL URLWithString:@"http://csdb.dk/release/download.php?id=125155"];
    
    _isInBackground = false;
    
    return YES;
}
							

// ----------------------------------------------------------------------------
- (void) applicationWillResignActive:(UIApplication *)application
// ----------------------------------------------------------------------------
{
    // Sent when the application is about to move from active to inactive state. This can occur for certain types of temporary interruptions (such as an incoming phone call or SMS message) or when the user quits the application and it begins the transition to the background state.
    // Use this method to pause ongoing tasks, disable timers, and throttle down OpenGL ES frame rates. Games should use this method to pause the game.
    
    _isInBackground = true;
    
    UIViewController* presentedViewController = [self.window.rootViewController presentedViewController];
    if (presentedViewController != nil)
    {
        EmulationViewController* emulationViewController = (EmulationViewController*)presentedViewController;
        [emulationViewController setIsInBackground:_isInBackground];
    }
}


// ----------------------------------------------------------------------------
- (void) applicationDidEnterBackground:(UIApplication *)application
// ----------------------------------------------------------------------------
{
    // Use this method to release shared resources, save user data, invalidate timers, and store enough application state information to restore your application to its current state in case it is terminated later. 
    // If your application supports background execution, this method is called instead of applicationWillTerminate: when the user quits.
}


// ----------------------------------------------------------------------------
- (void) applicationWillEnterForeground:(UIApplication *)application
// ----------------------------------------------------------------------------
{
    // Called as part of the transition from the background to the inactive state; here you can undo many of the changes made on entering the background.

    _isInBackground = false;
    
    UIViewController* presentedViewController = [self.window.rootViewController presentedViewController];
    if (presentedViewController != nil)
    {
        EmulationViewController* emulationViewController = (EmulationViewController*)presentedViewController;
        //NSLog(@"Root: %@", [emulationViewController description]);
        [emulationViewController setIsInBackground:_isInBackground];
    }
}


// ----------------------------------------------------------------------------
- (void) applicationDidBecomeActive:(UIApplication *)application
// ----------------------------------------------------------------------------
{
    // Restart any tasks that were paused (or not yet started) while the application was inactive. If the application was previously in the background, optionally refresh the user interface.
}


// ----------------------------------------------------------------------------
- (void) applicationWillTerminate:(UIApplication *)application
// ----------------------------------------------------------------------------
{
    // Called when the application is about to terminate. Save data if appropriate. See also applicationDidEnterBackground:.
}


// ----------------------------------------------------------------------------
- (BOOL)application:(UIApplication *)application openURL:(NSURL *)url sourceApplication:(NSString *)sourceApplication annotation:(id)annotation
// ----------------------------------------------------------------------------
{
	NSLog(@"URL: handle %@", [url absoluteString]);
    
    // Get right view of split view
    UISplitViewController* splitViewController = (UISplitViewController *)self.window.rootViewController;
    UIViewController* listViewController = [splitViewController.viewControllers lastObject];
    
    NSArray* pathComponents = [url pathComponents];
    NSString* type = [url host];
    NSString* identifier = [pathComponents objectAtIndex:1];
    
    if (type == nil)
    {
        UIViewController* presentedViewController = [self.window.rootViewController presentedViewController];
        if (presentedViewController != nil)
        {
            EmulationViewController* runningEmuViewController = (EmulationViewController*)presentedViewController;
            [runningEmuViewController clickDoneButton:nil];
            
            while ([runningEmuViewController isRunning])
            {
                [NSThread sleepForTimeInterval:0.1];
            }
                
            splitViewController = (UISplitViewController *)self.window.rootViewController;
            listViewController = [splitViewController.viewControllers lastObject];
        }

        UIStoryboard* storyBoard = [UIStoryboard storyboardWithName:INTERFACE_IS_PHONE ? @"iPhoneStoryboard" : @"MainStoryboard" bundle:nil];
        EmulationViewController* emulationViewController = [storyBoard instantiateViewControllerWithIdentifier:@"EmulationViewController"];
        
        emulationViewController.dataFileURLString = url.absoluteString;
        emulationViewController.title = url.lastPathComponent;
        emulationViewController.releaseId = @"";
        emulationViewController.releaseViewController = nil;
        
        [splitViewController presentViewController:emulationViewController animated:YES completion:NULL];
    }
    else
    {
        if ([type caseInsensitiveCompare:@"release"] == NSOrderedSame)
        {
            //NSLog(@"release %@", identifier);
            CSReleaseItem* releaseItem = (CSReleaseItem*) [[CSDatabaseCache sharedInstance] itemForType:CSDB_RELEASE withIdentifier:identifier];

            if (INTERFACE_IS_PHONE && [listViewController conformsToProtocol:@protocol(CSReleaseViewerDelegate)])
            {
                UIViewController<CSReleaseViewerDelegate>* viewController = (id)listViewController;
                viewController.selectedReleaseItem = releaseItem;
                [viewController performSegueWithIdentifier:@"ShowReleaseDetail" sender:self];
            }
            
            if (INTERFACE_IS_PAD)
            {
                UINavigationController* navigationController = (UINavigationController*)listViewController;
                if ([navigationController.topViewController conformsToProtocol:@protocol(CSReleaseViewerDelegate)])
                {
                    UIViewController<CSReleaseViewerDelegate>* rightViewController = (id)navigationController.topViewController;
                    rightViewController.selectedReleaseItem = releaseItem;
                    [rightViewController performSegueWithIdentifier:@"ShowReleaseDetail" sender:self];
                }
                
            }
            
        }
        else if ([type caseInsensitiveCompare:@"scener"] == NSOrderedSame)
        {
            //NSLog(@"scener %@", identifier);

            CSScenerItem* scenerItem = (CSScenerItem*) [[CSDatabaseCache sharedInstance] itemForType:CSDB_SCENER withIdentifier:identifier];

            if (INTERFACE_IS_PHONE && [listViewController conformsToProtocol:@protocol(CSScenerViewerDelegate)])
            {
                UIViewController<CSScenerViewerDelegate>* viewController = (id)listViewController;
                viewController.selectedScenerItem = scenerItem;
                [viewController performSegueWithIdentifier:@"ShowScenerDetail" sender:self];
            }
            
            if (INTERFACE_IS_PAD)
            {
                UINavigationController* navigationController = (UINavigationController*)listViewController;
                if ([navigationController.topViewController conformsToProtocol:@protocol(CSScenerViewerDelegate)])
                {
                    UIViewController<CSScenerViewerDelegate>* rightViewController = (id)navigationController.topViewController;
                    rightViewController.selectedScenerItem = scenerItem;
                    [rightViewController performSegueWithIdentifier:@"ShowScenerDetail" sender:self];
                }
            }
        }
    }
    
	return YES;
}

@end


@implementation UINavigationController (overrides)


// ----------------------------------------------------------------------------
- (BOOL) shouldAutorotate
// ----------------------------------------------------------------------------
{
    return [self.topViewController shouldAutorotate];
    
//    if (INTERFACE_IS_PHONE)
//    {
//        id currentViewController = self.topViewController;
//        
//        if ([currentViewController isKindOfClass:[ReleaseViewController class]] ||
//            [currentViewController isKindOfClass:[ScenerViewController class]] ||
//            [currentViewController isKindOfClass:[CreditsViewController class]] ||
//            [currentViewController isKindOfClass:[DetailViewController class]] ||
//            [currentViewController isKindOfClass:[MasterViewController class]] )
//            
//            return NO;
//    }
//    
//    return YES;
}
@end
