//
//  DetailViewController.m
//  C64SceneBrowser
//
//  Created by Andreas Varga on 16.03.12.
//  Copyright (c) 2012 __MyCompanyName__. All rights reserved.
//

#import "DetailViewController.h"

#import <QuartzCore/QuartzCore.h>

#import "UIImageView+WebCache.h"
#import "CSReleaseItem.h"
#import "CSDatabaseCache.h"
#import "SSLoadingView.h"
#import "ReleaseViewController.h"
#import "TBXML.h"
#import "AppDelegate.h"


@implementation DetailViewController

@synthesize overviewItem = _overviewItem;
@synthesize selectedReleaseItem = _selectedReleaseItem;
@synthesize isAppearing = _isAppearing;


#define BATCH_SIZE 25

#pragma mark - Managing the detail item


// ----------------------------------------------------------------------------
- (void)setOverviewItem:(OverviewItem*)overviewItem
// ----------------------------------------------------------------------------
{
    if (_overviewItem != overviewItem)
    {
        _overviewItem = overviewItem;
        
        // Update the view.
        [self configureView];
    }
}


// ----------------------------------------------------------------------------
- (void) configureView
// ----------------------------------------------------------------------------
{
    if (_overviewItem != nil)
    {
        self.navigationItem.title = _overviewItem.description;
        _viewMode = _overviewItem.type;
        self.navigationItem.rightBarButtonItem = nil;
        _gridView.editing = NO;
        _addLoadMoreItem = YES;
        _loadingMoreItems = NO;
        _releasesLoaded = 0;
        _isPortrait = UIInterfaceOrientationIsPortrait(self.interfaceOrientation);
        
        [_gridView scrollToObjectAtIndex:0 animated:NO];
        
        _items = [[NSMutableArray alloc] init];

        if (_loadingView == nil)
        {
            CGSize size = self.view.frame.size;
            _loadingView = [[SSLoadingView alloc] initWithFrame:CGRectMake(0.0f, 0.0f, size.width, size.height)];
            [self.view addSubview:_loadingView];
        }
            
        //NSLog(@"Adding request for release list\n");
        
        [_releaseListUrlConnection cancel];
        
        if (_overviewItem.url != nil)
        {
            NSString* url = _overviewItem.url;
            if (_overviewItem.hasRange)
                url = [url stringByAppendingFormat:@"&start=%d&count=%d", 0, BATCH_SIZE];

            _releasesLoaded = BATCH_SIZE;
                
            //NSLog(@"Overview URL: %@", url);
            NSMutableURLRequest* request = [NSMutableURLRequest requestWithURL:[NSURL URLWithString:url] cachePolicy:NSURLRequestUseProtocolCachePolicy timeoutInterval:60.0];
            _releaseListData = [NSMutableData data];
            _releaseListUrlConnection = [[NSURLConnection alloc] initWithRequest:request delegate:self];
        }
        else if (_overviewItem.type == RECENTLY_VIEWED)
        {
            NSString* releasesPath = [[CSDatabaseCache sharedInstance] releasesPath];
            NSArray* releaseDirNames = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:releasesPath error:NULL];

            NSMutableDictionary* releaseIdForTimeStamp = [NSMutableDictionary dictionaryWithCapacity:releaseDirNames.count];

            // For each release directory, get its path and its time stamp, and add them to a dictionary, with the timestamp as key
            for (NSString* releaseDirName in releaseDirNames)
            {
                NSString* releasePath = [releasesPath stringByAppendingPathComponent:releaseDirName];
                
                BOOL is_directory = NO;
                [[NSFileManager defaultManager] fileExistsAtPath:releasePath isDirectory:&is_directory];
                if (is_directory)
                {
                    NSTimeInterval time = [[CSDatabaseCache sharedInstance] readTimeStamp:releaseDirName ofType:CSDB_RELEASE];
                    NSNumber* timeNumber = [NSNumber numberWithDouble:time];
                    [releaseIdForTimeStamp setObject:releaseDirName forKey:timeNumber];
                }
            }
            
            // Sort the releases by time stamp
            NSArray* sortedKeys = [[releaseIdForTimeStamp allKeys] sortedArrayUsingSelector:@selector(compare:)];
            _recentlyViewedReleaseIds = [NSMutableArray arrayWithCapacity:sortedKeys.count];
            
            // Get the first N release items from the sorted array
            NSMutableArray* firstBatchReleaseIds = [NSMutableArray arrayWithCapacity:BATCH_SIZE];
            for (NSNumber* sortedKey in [sortedKeys reverseObjectEnumerator])
            {
                NSString* releaseId = [releaseIdForTimeStamp objectForKey:sortedKey];
                [_recentlyViewedReleaseIds addObject:releaseId];
                if (firstBatchReleaseIds.count < BATCH_SIZE)
                    [firstBatchReleaseIds addObject:releaseId];
            }
            
            _addLoadMoreItem = _recentlyViewedReleaseIds.count > BATCH_SIZE;
            
            [self loadReleasesIds:firstBatchReleaseIds];
            _releasesLoaded = BATCH_SIZE;
            
            _noRecentsView.hidden = firstBatchReleaseIds.count != 0;
            _gridView.hidden = firstBatchReleaseIds.count == 0;
        }
        else if (_overviewItem.type == FAVORITES)
        {
            _addLoadMoreItem = NO;
            
            NSMutableArray* releaseIds = [[CSDatabaseCache sharedInstance] favorites];
            [self loadReleasesIds:releaseIds];

            _noFavoritesView.hidden = releaseIds.count != 0;
            _gridView.hidden = releaseIds.count == 0;
            
            self.navigationItem.rightBarButtonItem = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemEdit target:self action:@selector(clickEditButton:)];
        }
    }
}



// ----------------------------------------------------------------------------
- (void) loadReleasesIds:(NSMutableArray*)releaseIds
// ----------------------------------------------------------------------------
{
    if (releaseIds.count == 0)
    {
        [_loadingView removeFromSuperview];
        _loadingView = nil;
        return;
    }
 
    // Remove load more item if it was there
    CSReleaseItem* lastItem = [_items lastObject];
    if (lastItem.is_loadmore_item)
        [_items removeLastObject];
    
    // Add an item for each release
    for (NSString* releaseId in releaseIds)
    {
        CSReleaseItem* item = (CSReleaseItem*)[[CSDatabaseCache sharedInstance] itemForType:CSDB_RELEASE withIdentifier:releaseId];
        [_items addObject:item];
    }
    
    // Find first item that needs caching
    CSReleaseItem* item = nil;
    for (item in _items)
    {
        if (item.cached_depth < 1)
            break;
    }

    if (item != nil)
        [item cacheAndNotify:self withDepth:1];
    
    [_gridView reloadData];
    [_loadingView removeFromSuperview];
    _loadingView = nil;
}


// ----------------------------------------------------------------------------
- (void) loadReleaseItems:(NSArray*)releaseItems
// ----------------------------------------------------------------------------
{
    // Remove load more item if it was there
    CSReleaseItem* lastItem = [_items lastObject];
    if (lastItem.is_loadmore_item)
        [_items removeLastObject];
    
    // Append all items from list
    [_items addObjectsFromArray:releaseItems];

    if (_loadingView != nil)
    {
        [_loadingView removeFromSuperview];
        _loadingView = nil;
    }
    
    // Add special load more item
    if (_addLoadMoreItem)
    {
        CSReleaseItem* loadmore_item = [[CSReleaseItem alloc] init];
        loadmore_item.is_loadmore_item = YES;
        [_items addObject:loadmore_item];
    }

    [_gridView reloadData];
}



#pragma mark - CSDatabaseItemDelegate protocol


// ----------------------------------------------------------------------------
- (void) databaseItemIsCached:(CSDatabaseItem*)item
// ----------------------------------------------------------------------------
{
    NSUInteger cached_index = [_items indexOfObject:item];
    if (cached_index == NSNotFound)
        return;

    NSUInteger next_index = cached_index + 1;
    if (next_index < _items.count)
    {
        // Cache next item
        CSReleaseItem* next_item = [_items objectAtIndex:next_index];
        while (next_item != nil && next_item.cached_depth >= 1)
        {
            next_index++;
            if (next_index < _items.count)
                next_item = [_items objectAtIndex:next_index];
            else
                break;
        }
            
        if (next_item != nil && next_index < _items.count)
            [next_item cacheAndNotify:self withDepth:1];
    }
    else if (_addLoadMoreItem)
    {
        // Special load more item
        CSReleaseItem* loadmore_item = [[CSReleaseItem alloc] init];
        loadmore_item.is_loadmore_item = YES;
        [_items addObject:loadmore_item];
    }
    
    [_gridView reloadData];
}


// ----------------------------------------------------------------------------
- (void) cachingFailed
// ----------------------------------------------------------------------------
{
    [[CSDatabaseCache sharedInstance] showNetworkError];
}


#pragma mark -

// ----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didReceiveResponse:(NSURLResponse*)response
// ----------------------------------------------------------------------------
{
    if (connection == _releaseListUrlConnection)
    {
        [_releaseListData setLength:0];
        return;
    }
}


// ----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didReceiveData:(NSData*)data
// ----------------------------------------------------------------------------
{
    if (connection == _releaseListUrlConnection)
    {
        [_releaseListData appendData:data];
        return;
    }
}


// ----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didFailWithError:(NSError*)error
// ----------------------------------------------------------------------------
{
    if (_loadingView != nil)
    {
        [_loadingView removeFromSuperview];
        _loadingView = nil;
    }

    _loadingMoreItems = NO;
    [_items removeAllObjects];
    [_gridView reloadData];
    
    [[CSDatabaseCache sharedInstance] showNetworkError];
}


// ----------------------------------------------------------------------------
- (void) connectionDidFinishLoading:(NSURLConnection*)connection
// ----------------------------------------------------------------------------
{
    if (connection == _releaseListUrlConnection && [_releaseListData length] > 0)
    {
        //NSLog(@"Release list download finished\n");
        
        _loadingMoreItems = NO;

        if (!_overviewItem.parseHTML)
        {
            NSMutableArray* releaseItems = [NSMutableArray array];
            
            NSError* error;
            TBXML* tbxml = [TBXML tbxmlWithXMLData:_releaseListData error:&error];
            if (tbxml == NULL || tbxml.rootXMLElement == NULL)
                return;
            
            TBXMLElement* releaseListElement = [TBXML childElementNamed:@"CSDbLatestReleases" parentElement:tbxml.rootXMLElement];
            if (releaseListElement == NULL)
                releaseListElement = [TBXML childElementNamed:@"CSDbLatestAdditions" parentElement:tbxml.rootXMLElement];
            if (releaseListElement == NULL)
                releaseListElement = [TBXML childElementNamed:@"CSDbSearchResult" parentElement:tbxml.rootXMLElement];
            if (releaseListElement == NULL)
                releaseListElement = [TBXML childElementNamed:@"CSDbChart" parentElement:tbxml.rootXMLElement];
            
            if (releaseListElement != NULL)
            {
                TBXMLElement* releaseElement = [TBXML childElementNamed:@"Release" parentElement:releaseListElement];
                while (releaseElement != NULL)
                {
                    NSString* releaseId = [TBXML textForElement:[TBXML childElementNamed:@"ID" parentElement:releaseElement] withStringEncoding:NSUTF8StringEncoding];
                    
                    CSReleaseItem* releaseItem = (CSReleaseItem*)[[CSDatabaseCache sharedInstance] itemForType:CSDB_RELEASE withIdentifier:releaseId];
                    [releaseItem setFromXMLElement:releaseElement];
                    releaseItem.cached_depth = 1;
                    [releaseItems addObject:releaseItem];
                    
                    releaseElement = [TBXML nextSiblingNamed:@"Release" searchFromElement:releaseElement];
                }

                TBXMLElement* entryElement = [TBXML childElementNamed:@"Entry" parentElement:releaseListElement];
                while (entryElement != NULL)
                {
                    TBXMLElement* releaseElement = [TBXML childElementNamed:@"Release" parentElement:entryElement];
                    if (releaseElement != NULL)
                    {
                        NSString* releaseId = [TBXML textForElement:[TBXML childElementNamed:@"ID" parentElement:releaseElement] withStringEncoding:NSUTF8StringEncoding];
                        
                        CSReleaseItem* releaseItem = (CSReleaseItem*)[[CSDatabaseCache sharedInstance] itemForType:CSDB_RELEASE withIdentifier:releaseId];
                        [releaseItem setFromXMLElement:releaseElement];
                        releaseItem.cached_depth = 1;
                        [releaseItems addObject:releaseItem];
                    }
                    
                    entryElement = [TBXML nextSiblingNamed:@"Entry" searchFromElement:entryElement];
                }
            }
            
            if (_overviewItem.type == SEARCH)
            {
                NSInteger results = releaseItems.count;
                
                TBXMLElement* infoElement = [TBXML childElementNamed:@"Info" parentElement:releaseListElement];
                if (infoElement != NULL)
                {
                    NSString* infoString = [TBXML textForElement:infoElement];
                    if ([infoString caseInsensitiveCompare:@"No result"] == NSOrderedSame)
                    {
                        results = 0;
                    }
                    else
                    {
                        NSArray* components = [infoString componentsSeparatedByString:@" "];
                        results = [[components objectAtIndex:6] integerValue];
                    }
                }
                
                //NSLog(@"Search results: %d", results);

                _addLoadMoreItem = results > _releasesLoaded;

                self.navigationItem.title = [NSString stringWithFormat:@"%@ (%ld results)", _overviewItem.description, (long)results];
            }
            
            [self loadReleaseItems:releaseItems];
            
//            if (_overviewItem.type == SEARCH && _releasesLoaded == BATCH_SIZE && _items.count == 1)
//                [self viewItemAtIndex:0];
            
            return;
        }
        
        NSString* releases_string = [[NSString alloc] initWithData:_releaseListData encoding:NSASCIIStringEncoding];
        
        NSMutableArray* releaseIds = [NSMutableArray array];
        
        NSRange result_range = NSMakeRange(0, 0);
        
        while (result_range.location != NSNotFound)
        {
            NSUInteger start_location = result_range.location + result_range.length;
            NSRange search_range = NSMakeRange(start_location, [releases_string length] - start_location);
            result_range = [releases_string rangeOfString:@"release/?id=" options:NSLiteralSearch range:search_range];
            if (result_range.location != NSNotFound)
            {
                NSRange search_range2 = NSMakeRange(result_range.location + result_range.length, 100);
                NSRange result_range2 = [releases_string rangeOfString:@"&show=" options:NSLiteralSearch range:search_range2];
                if (result_range2.location == NSNotFound)
                    result_range2 = [releases_string rangeOfString:@"\">" options:NSLiteralSearch range:search_range2];

                if (result_range2.location != NSNotFound)
                {
                    NSUInteger id_location = result_range.location + result_range.length;
                    NSRange id_range = NSMakeRange(id_location, result_range2.location - id_location);
                    NSString* release_id = [releases_string substringWithRange:id_range];
                    //NSLog(@"release: %@\n", release_id);
                    
                    if (![releaseIds containsObject:release_id])
                        [releaseIds addObject:release_id];
                }
            }

            if ([releaseIds count] == BATCH_SIZE)
                break;
        }

        //NSLog(@"Starting to cache releases\n");
        
        [self loadReleasesIds:releaseIds];
    }
}


// ----------------------------------------------------------------------------
- (void) loadMoreReleases
// ----------------------------------------------------------------------------
{
    // Load more item
    if (_overviewItem.hasRange)
    {
        NSString* url = [_overviewItem.url stringByAppendingFormat:@"&start=%d&count=%d", _releasesLoaded, BATCH_SIZE];
        //NSLog(@"Load more URL: %@", url);
        
        NSMutableURLRequest* request = [NSMutableURLRequest requestWithURL:[NSURL URLWithString:url] cachePolicy:NSURLRequestUseProtocolCachePolicy timeoutInterval:60.0];
        _releaseListData = [NSMutableData data];
        _releaseListUrlConnection = [[NSURLConnection alloc] initWithRequest:request delegate:self];
        
        _loadingMoreItems = YES;
        _releasesLoaded += BATCH_SIZE;
        [_gridView reloadData];
    }
    else if (_overviewItem.type == RECENTLY_VIEWED)
    {
        NSMutableArray* nextBatchReleaseIds = [NSMutableArray arrayWithCapacity:BATCH_SIZE];
        _addLoadMoreItem = _recentlyViewedReleaseIds.count > _releasesLoaded+BATCH_SIZE;
        for (int i=_releasesLoaded; i<_recentlyViewedReleaseIds.count; i++)
        {
            NSString* releaseId = [_recentlyViewedReleaseIds objectAtIndex:i];
            if (nextBatchReleaseIds.count < BATCH_SIZE)
                [nextBatchReleaseIds addObject:releaseId];
            else
                break;
        }
        
        [self loadReleasesIds:nextBatchReleaseIds];
        _releasesLoaded += BATCH_SIZE;
    }
}


#pragma mark - ViewController events

// ----------------------------------------------------------------------------
- (void) viewDidLoad
// ----------------------------------------------------------------------------
{
    [super viewDidLoad];

    if (_loadingView != nil)
        [_loadingView removeFromSuperview];
    _loadingView = nil;
    _releasesLoaded = 0;
    _isAppearing = NO;
}


// ----------------------------------------------------------------------------
- (void) viewDidUnload
// ----------------------------------------------------------------------------
{
    _gridView = nil;
    _items = nil;
    _loadingView = nil;

    [super viewDidUnload];
}





// ----------------------------------------------------------------------------
- (BOOL) shouldAutorotate
// ----------------------------------------------------------------------------
{
    return YES;
}


// ----------------------------------------------------------------------------
- (void)viewWillTransitionToSize:(CGSize)size withTransitionCoordinator:(id<UIViewControllerTransitionCoordinator>)coordinator
// ----------------------------------------------------------------------------
{
    [super viewWillTransitionToSize:size withTransitionCoordinator:coordinator];
    
    _isPortrait = size.width < size.height;
    
    [_gridView reloadData];
}


// ----------------------------------------------------------------------------
- (void) loadView 
// ----------------------------------------------------------------------------
{
    [super loadView];
    self.view.backgroundColor = [UIColor clearColor];
    
    NSInteger spacing = INTERFACE_IS_PHONE ? 10 : 15;
    
    GMGridView* gridView = [[GMGridView alloc] initWithFrame:self.view.bounds];
    gridView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
#ifdef BLACK_MODE
    gridView.backgroundColor = [UIColor colorWithWhite:0.1f alpha:1.0f]; 
    //gridView.backgroundColor = [UIColor blackColor];
#else
    gridView.backgroundColor = [UIColor clearColor];
#endif
    [self.view addSubview:gridView];
    
    _gridView = gridView;
    
    _gridView.style = GMGridViewStyleSwap;
    _gridView.itemSpacing = spacing;
    _gridView.minEdgeInsets = UIEdgeInsetsMake(spacing, spacing, spacing, spacing);
    _gridView.centerGrid = YES;
    _gridView.actionDelegate = self;
    _gridView.sortingDelegate = nil;
    _gridView.transformDelegate = nil;
    _gridView.dataSource = self;
}


// ----------------------------------------------------------------------------
- (void) viewWillAppear:(BOOL)animated
// ----------------------------------------------------------------------------
{
    _isAppearing = YES;
    
    [super viewWillAppear:animated];
    
    UIViewController* masterViewController = [[self.splitViewController.viewControllers objectAtIndex:0] topViewController];
    [masterViewController.navigationController popViewControllerAnimated:animated];

    if (self.toolbarItems.count == 0)
    {
        AppDelegate* appDelegate = (AppDelegate*)[[UIApplication sharedApplication] delegate];
        if (appDelegate.masterPopoverController != nil)
            [self addMasterPopoverButton:appDelegate.masterPopoverBarButtonItem];
    }
    else
    {
        UIBarButtonItem* barButtonItem = [self.toolbarItems objectAtIndex:0];
        barButtonItem.title = @"Categories";
    }
}


// ----------------------------------------------------------------------------
- (void) viewDidAppear:(BOOL)animated
// ----------------------------------------------------------------------------
{
    _isAppearing = NO;
}


// ----------------------------------------------------------------------------
- (void) addMasterPopoverButton:(UIBarButtonItem*)barButtonItem
// ----------------------------------------------------------------------------
{
    barButtonItem.title = @"Categories";
    NSMutableArray* toolbarItems = [NSMutableArray arrayWithObject:barButtonItem];
    [self setToolbarItems:[toolbarItems arrayByAddingObjectsFromArray:self.toolbarItems] animated:YES];
}


// ----------------------------------------------------------------------------
- (void) removeMasterPopoverButton:(UIBarButtonItem*)barButtonItem
// ----------------------------------------------------------------------------
{
    if (self.toolbarItems.count > 0 && [self.toolbarItems objectAtIndex:0] == barButtonItem)
        [self setToolbarItems:[self.toolbarItems subarrayWithRange:NSMakeRange(1, self.toolbarItems.count-1)] animated:YES];
}


// ----------------------------------------------------------------------------
- (void) clickEditButton:(id)sender
// ----------------------------------------------------------------------------
{
    _gridView.editing = YES;
    self.navigationItem.rightBarButtonItem = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemDone target:self action:@selector(clickDoneButton:)];
}


// ----------------------------------------------------------------------------
- (void) clickDoneButton:(id)sender
// ----------------------------------------------------------------------------
{
    _gridView.editing = NO;
    self.navigationItem.rightBarButtonItem = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemEdit target:self action:@selector(clickEditButton:)];
}


// ----------------------------------------------------------------------------
- (void) viewItemAtIndex:(NSInteger)index
// ----------------------------------------------------------------------------
{
    _selectedReleaseItem = [_items objectAtIndex:index];
    
    if (_selectedReleaseItem.cached_depth > 0)
        [self performSegueWithIdentifier:@"ShowReleaseDetail" sender:self];
    
}


#pragma mark - GMGridViewDataSource


// ----------------------------------------------------------------------------
- (NSInteger) numberOfItemsInGMGridView:(GMGridView *)gridView
// ----------------------------------------------------------------------------
{
    return [_items count];
}


// ----------------------------------------------------------------------------
- (CGSize) sizeForItemsInGMGridView:(GMGridView *)gridView
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE)
    {
        if (_isPortrait)
            return CGSizeMake(300, 244);
        else
            return CGSizeMake(260, 212);
    }
    else
        return CGSizeMake(210, 160);
}


// ----------------------------------------------------------------------------
- (GMGridViewCell*) GMGridView:(GMGridView *)gridView cellForItemAtIndex:(NSInteger)index
// ----------------------------------------------------------------------------
{
    CGSize size = [self sizeForItemsInGMGridView:gridView];
    GMGridViewCell* cell = [gridView dequeueReusableCell];
    if (cell == nil) 
    {
        cell = [[GMGridViewCell alloc] init];
        
        UIView* view = [[UIView alloc] initWithFrame:CGRectMake(0, 0, size.width, size.height)];
#ifdef BLACK_MODE
        view.backgroundColor = [UIColor colorWithWhite:0.15f alpha:1.0f];
#endif
        cell.contentView = view;
        cell.deleteButtonIcon = [UIImage imageNamed:@"delete_icon"];
        cell.deleteButtonOffset = CGPointMake(-14.0f, -14.0f);
    }

    [[cell.contentView subviews] makeObjectsPerformSelector:@selector(removeFromSuperview)];
    
    CGRect imageRect;
    
    if (INTERFACE_IS_PHONE)
        imageRect = CGRectMake(2.0f, 5.0f, size.width - 4.0f, size.height - 40.0f);
    else
        imageRect = CGRectMake(0.0f, 5.0f, size.width, size.height - 40.0f);
    UIImageView* imageView = [[UIImageView alloc] initWithFrame:imageRect];

    CSReleaseItem* release_item = [_items objectAtIndex:index];

    if (release_item.is_loadmore_item)
    {
        imageView.image = nil;
        if (INTERFACE_IS_PHONE)
            imageView.contentMode = UIViewContentModeScaleAspectFit;
        else
            imageView.contentMode = UIViewContentModeCenter;
        [cell.contentView addSubview:imageView];
        
        UILabel *loadMoreLabel = [[UILabel alloc] initWithFrame:imageView.frame];
        loadMoreLabel.text = @"Load more...";
        loadMoreLabel.textAlignment = NSTextAlignmentCenter;
        loadMoreLabel.backgroundColor = [UIColor clearColor];
#ifdef BLACK_MODE
        loadMoreLabel.textColor = [UIColor whiteColor];
#endif
        loadMoreLabel.font = [UIFont systemFontOfSize:22.0f];
        [cell.contentView addSubview:loadMoreLabel];
        
        if (_loadingMoreItems)
        {
            loadMoreLabel.text = @"Loading...";
            
            UIActivityIndicatorView* activityView = [[UIActivityIndicatorView alloc] initWithActivityIndicatorStyle:UIActivityIndicatorViewStyleWhite];
            activityView.frame = CGRectMake(INTERFACE_IS_PHONE ? (size.width - 222.0f) : 18.0f, 5.0f, 16.0f, size.height - 40.0f);
            [activityView startAnimating];
            [cell.contentView addSubview:activityView];
        }
    }
    else
    {
        BOOL has_screenshot = YES;
        
        //NSLog(@"Creating view index %d, name: %@", index, item.name);
        if (release_item.cached_depth > 0)
        {
            if (release_item.screenshot != nil && release_item.screenshot.length > 0)
            {
                NSURL* url = [NSURL URLWithString:release_item.screenshot];
                [imageView setImageWithURL:url placeholderImage:[UIImage imageNamed:@"placeholder"]];
            }
            else
            {
                imageView.image = nil;
                has_screenshot = NO;
            }
        }
        else
            imageView.image = nil;
        
        imageView.contentMode = UIViewContentModeScaleAspectFit;
        [cell.contentView addSubview:imageView];
        
        if (!has_screenshot)
        {
            UILabel *noScreenshotLabel = [[UILabel alloc] initWithFrame:CGRectMake(10.0f, 5.0f, size.width - 20.0f, size.height - 40.0f)];
            
            noScreenshotLabel.text = @"No screenshot available";
            noScreenshotLabel.textAlignment = NSTextAlignmentCenter;
            noScreenshotLabel.lineBreakMode = NSLineBreakByWordWrapping;
            noScreenshotLabel.numberOfLines = 0;
            noScreenshotLabel.backgroundColor = [UIColor clearColor];
#ifdef BLACK_MODE
            noScreenshotLabel.textColor = [UIColor whiteColor];
#endif
            noScreenshotLabel.font = [UIFont systemFontOfSize:22.0f];
            [cell.contentView addSubview:noScreenshotLabel];
        }
        
        UILabel *titleLabel = [[UILabel alloc] initWithFrame:CGRectMake(5.0f, size.height - 35.0f, size.width - 10.0f, 21.0f)];
        
        if (release_item.name != nil)
            titleLabel.text = release_item.name;
        else
            titleLabel.text = @"";
        titleLabel.textAlignment = NSTextAlignmentLeft;
        titleLabel.backgroundColor = [UIColor clearColor];
#ifdef BLACK_MODE
        titleLabel.textColor = [UIColor whiteColor];
#endif
        titleLabel.font = [UIFont boldSystemFontOfSize:12];
        titleLabel.adjustsFontSizeToFitWidth = YES;
        [cell.contentView addSubview:titleLabel];
        
        UILabel *typeLabel = [[UILabel alloc] initWithFrame:CGRectMake(5.0f, size.height - 21.0f, size.width * 0.58f, 21.0f)];
        //UILabel *typeLabel = [[UILabel alloc] initWithFrame:CGRectMake(5.0f, 139.0f, 120.0f, 21.0f)];

        if (release_item.releaseType != nil)
            typeLabel.text = release_item.releaseType;
        else
            typeLabel.text = @"";
        typeLabel.textAlignment = NSTextAlignmentLeft;
        typeLabel.backgroundColor = [UIColor clearColor];
#ifdef BLACK_MODE
        typeLabel.textColor = [UIColor whiteColor];
#endif
        typeLabel.font = [UIFont systemFontOfSize:11];
        [cell.contentView addSubview:typeLabel];
        
        UILabel *dateLabel = [[UILabel alloc] initWithFrame:CGRectMake(size.width - size.width * 0.4f, size.height - 21.0f, size.width * 0.38f, 21.0f)];
        //UILabel *dateLabel = [[UILabel alloc] initWithFrame:CGRectMake(125.0f, 139.0f, 80.0f, 21.0f)];
        
        
        NSString* date = [[CSDatabaseCache sharedInstance] stringForDate:release_item.releaseDate short:YES];
        if (date != nil)
            dateLabel.text = date;
        else
            dateLabel.text = @"";
        
        dateLabel.textAlignment = NSTextAlignmentRight;
        dateLabel.backgroundColor = [UIColor clearColor];
#ifdef BLACK_MODE
        dateLabel.textColor = [UIColor whiteColor];
#endif
        dateLabel.font = [UIFont systemFontOfSize:11];
        [cell.contentView addSubview:dateLabel];
        
        /*
         UILabel *idLabel = [[UILabel alloc] initWithFrame:CGRectMake(0.0f, 0.0f, 103.0f, 12.0f)];
         idLabel.text = release_item.identifier;
         idLabel.textAlignment = NSTextAlignmentLeft;
         idLabel.backgroundColor = [UIColor clearColor];
         idLabel.textColor = [UIColor whiteColor];
         idLabel.font = [UIFont systemFontOfSize:11];
         [cell.contentView addSubview:idLabel];
         
         UILabel *indexLabel = [[UILabel alloc] initWithFrame:CGRectMake(100.0f, 0.0, 100.0f, 12.0f)];
         indexLabel.text = [NSString stringWithFormat:@"%d", index];
         indexLabel.textAlignment = NSTextAlignmentRight;
         indexLabel.backgroundColor = [UIColor clearColor];
         indexLabel.textColor = [UIColor whiteColor];
         indexLabel.font = [UIFont systemFontOfSize:11];
         [cell.contentView addSubview:indexLabel];
         */
    }
     
    return cell;
}


// ----------------------------------------------------------------------------
- (void) GMGridView:(GMGridView *)gridView deleteItemAtIndex:(NSInteger)index
// ----------------------------------------------------------------------------
{
    CSReleaseItem* deletedItem = [_items objectAtIndex:index];
    [[CSDatabaseCache sharedInstance] removeFavorite:deletedItem.identifier];
    
    [_items removeObjectAtIndex:index];
}

#pragma mark - GMGridViewActionDelegate


// ----------------------------------------------------------------------------
- (void) GMGridView:(GMGridView *)gridView didTapOnItemAtIndex:(NSInteger)index
// ----------------------------------------------------------------------------
{
    CSReleaseItem* item = [_items objectAtIndex:index];
    
    if (item.is_loadmore_item)
    {
        [self loadMoreReleases];
        return;
    }
    
    
    [self viewItemAtIndex:index];
}


// ----------------------------------------------------------------------------
- (void)GMGridView:(GMGridView *)gridView didScrollPastBottom:(float)distance
// ----------------------------------------------------------------------------
{
    if (!_loadingMoreItems)
        [self loadMoreReleases];
}


// ----------------------------------------------------------------------------
- (void)prepareForSegue:(UIStoryboardSegue *)segue sender:(id)sender
// ----------------------------------------------------------------------------
{
    if ([[segue identifier] isEqualToString:@"ShowReleaseDetail"])
    {
        ReleaseViewController* releaseViewController = [segue destinationViewController];
        releaseViewController.releaseItem = _selectedReleaseItem;
        
        if (self.toolbarItems.count == 1)
        {
            NSMutableArray* toolbarItems = [NSMutableArray arrayWithObject:[self.toolbarItems objectAtIndex:0]];
            releaseViewController.toolbarItems = [toolbarItems arrayByAddingObjectsFromArray:releaseViewController.toolbarItems];
        }
    }
}


#pragma mark - GMGridViewSortingDelegate


// ----------------------------------------------------------------------------
- (BOOL) GMGridView:(GMGridView *)gridView shouldAllowShakingBehaviorWhenMovingCell:(GMGridViewCell *)cell atIndex:(NSInteger)index
// ----------------------------------------------------------------------------
{
    return NO;
}


#pragma mark - DraggableGridViewTransformingDelegate


@end

