//
//  EmulationViewController.m
//  C64SceneBrowser
//
//  Created by Andreas Varga on 03.04.12.
//  Copyright (c) 2012 __MyCompanyName__. All rights reserved.
//

#import "EmulationViewController.h"
#import "SSZipArchive.h"
#import "SSLoadingView.h"
#import "CSDatabaseCache.h"
#import "ReleaseViewController.h"
#import "UIImage+GIF.h"
#import <GameController/GameController.h>
#import "OscilloscopeView.h"
#import "libx64/vice/src/arch/unix/ios/cocoa_touch/vicemachine.h"
#import "libx64/vice/src/arch/unix/ios/cocoa_touch/viceglview.h"
#import "libx64/vice/src/arch/unix/ios/cocoa_touch/vicenotifications.h"
extern "C" {
    #import "libx64/vice/src/arch/unix/ios/joy.h"
}

@implementation EmulationViewController


@synthesize dataFileURLString = _dataFileURLString;
@synthesize releaseId = _releaseId;
@synthesize joystickMode = _joystickMode;
@synthesize releaseViewController = _releaseViewController;
@synthesize keyboardVisible = _keyboardVisible;

static const float sControlsFadeDelay = 5.0f;

extern int c64scene_fake_vsid;

// ----------------------------------------------------------------------------
- (id) initWithNibName:(NSString *)nibNameOrNil bundle:(NSBundle *)nibBundleOrNil
// ----------------------------------------------------------------------------
{
    self = [super initWithNibName:nibNameOrNil bundle:nibBundleOrNil];
    if (self) 
    {
        
    }
    return self;
}


// ----------------------------------------------------------------------------
- (void) viewDidLoad
// ----------------------------------------------------------------------------
{
    [super viewDidLoad];
    
    _controlFadeTimer = nil;
    _controlsVisible = YES;
    
    _settingsPopOver = nil;
    _supportedFiles = nil;
    
    _emulationRunning = NO;
    _keyboardVisible = NO;
    _keyboardOffset = 0.0f;
    _viceViewScaled = NO;
    
    _gameController = nil;

    NSArray* controllers = [GCController controllers];
    if ([controllers count] > 0)
    {
        _gameController = [controllers objectAtIndex:0];
        [self fadeOutVirtualJoystickControls];
        [self setupGameController];
    }
        
    [self controlsFadeIn];
    
    if (_loadingView == nil)
    {
        CGSize size = self.view.frame.size;
        _loadingView = [[SSLoadingView alloc] initWithFrame:CGRectMake(0.0f, 0.0f, size.width, size.height)];
        [self.view addSubview:_loadingView];
    }
    
    _volumeView.showsVolumeSlider = YES;
    _volumeView2.showsVolumeSlider = YES;
    
    //[_volumeView setRouteButtonImage:[UIImage imageNamed:@"airplay"] forState:UIControlStateNormal];
    //[_volumeView setVolumeThumbImage:[UIImage imageNamed:@"volume_thumb"] forState:UIControlStateNormal];
    
//    for (UIView* view in [_volumeView subviews])
//    {
//        if ([view isKindOfClass:[UISlider class]])
//        {
//            ((UISlider*)view).minimumTrackTintColor = [UIColor whiteColor];
//            ((UISlider*)view).maximumTrackTintColor = [UIColor darkGrayColor];
//        }
//    }
    
    [self registerForNotifications];
    
    _titleLabel.text = self.title;
    
    _joystickMode = JOYSTICK_DISABLED;
    _joystickView.alpha = 0.0f;
    _fireButtonView.alpha = 0.0f;

    _driveLedView.hidden = YES;

    _toolBar.delegate = self;
    
    [self setNeedsStatusBarAppearanceUpdate];
    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(warpStatusNotification:)
                                                 name:@"WarpStatus"
                                               object:nil];
    
    [self startDataDownload];
}


// ----------------------------------------------------------------------------
- (void) viewDidUnload
// ----------------------------------------------------------------------------
{
    _viceView = nil;
    
    [super viewDidUnload];
}


// ----------------------------------------------------------------------------
- (void) viewWillAppear:(BOOL)animated
// ----------------------------------------------------------------------------
{
    [super viewWillAppear:animated];

    if (INTERFACE_IS_PHONE)
    {
        if (UIInterfaceOrientationIsPortrait(self.interfaceOrientation))
        {
            _bottomToolBar.hidden = YES;
            _portraitBottomToolBarView.hidden = NO;
        }
        else if (UIInterfaceOrientationIsLandscape(self.interfaceOrientation))
        {
            _bottomToolBar.hidden = NO;
            _portraitBottomToolBarView.hidden = YES;
        }
    }
    else
        _portraitBottomToolBarView.hidden = YES;
    
}


// ----------------------------------------------------------------------------
- (void) viewWillDisappear:(BOOL)animated
// ----------------------------------------------------------------------------
{
    [super viewWillDisappear:animated];
    
    [self cancelControlFade];
}


// ----------------------------------------------------------------------------
- (void) controllerWasConnected:(NSNotification*) notification
// ----------------------------------------------------------------------------
{
    // a controller was connected
    GCController* controller = (GCController*) notification.object;
    //NSLog(@"Controller connected: %@\n", controller.vendorName);
    
    _gameController = controller;
    [self fadeOutVirtualJoystickControls];
    
    [self setupGameController];
}


// ----------------------------------------------------------------------------
- (void) controllerWasDisconnected:(NSNotification*) notification
// ----------------------------------------------------------------------------
{
    // a controller was disconnected
    GCController* controller = (GCController*) notification.object;
    //NSLog(@"Controller disconnected: %@\n", controller.vendorName);
    
    _gameController = nil;
    [self fadeInVirtualJoystickControls];
}


static BOOL sIsUpPressed = NO;

// ----------------------------------------------------------------------------
- (void) setupGameController
// ----------------------------------------------------------------------------
{
    if (_gameController == nil)
        return;
    
    // register block for input change detection
    GCExtendedGamepad* profile = _gameController.extendedGamepad;

    profile.valueChangedHandler = ^(GCExtendedGamepad* gamepad, GCControllerElement* element)
    {
        if (element == gamepad.buttonA)
        {
            if (gamepad.buttonA.isPressed)
                sSetJoystickBit(16);
            else
                sClearJoystickBit(16);
        }

        if (element == gamepad.buttonB)
        {
            if (gamepad.buttonB.isPressed)
                [theVICEMachine setKeyRow:7 column:4 shift:0];
        }
        
        if (element == gamepad.buttonY)
        {
            if (gamepad.buttonY.isPressed)
            {
                sSetJoystickBit(1);
                sIsUpPressed = YES;
            }
            else
            {
                sClearJoystickBit(1);
                sIsUpPressed = NO;
            }
        }
        
        if (element == gamepad.dpad)
        {
            if (gamepad.dpad.up.isPressed)
                sSetJoystickBit(1);
            else if (!sIsUpPressed)
                sClearJoystickBit(1);

            if (gamepad.dpad.down.isPressed)
                sSetJoystickBit(2);
            else
                sClearJoystickBit(2);

            if (gamepad.dpad.left.isPressed)
                sSetJoystickBit(4);
            else
                sClearJoystickBit(4);

            if (gamepad.dpad.right.isPressed)
                sSetJoystickBit(8);
            else
                sClearJoystickBit(8);
        }
            
        
        if (element == gamepad.leftThumbstick)
        {
            if (gamepad.leftThumbstick.up.isPressed)
                sSetJoystickBit(1);
            else if (!sIsUpPressed)
                sClearJoystickBit(1);

            if (gamepad.leftThumbstick.down.isPressed)
                sSetJoystickBit(2);
            else
                sClearJoystickBit(2);

            if (gamepad.leftThumbstick.left.isPressed)
                sSetJoystickBit(4);
            else
                sClearJoystickBit(4);

            if (gamepad.leftThumbstick.right.isPressed)
                sSetJoystickBit(8);
            else
                sClearJoystickBit(8);
        }
        
        sUpdateJoystickState(_joystickMode - 1);
    };
    
    __weak typeof(self) weakSelf = self;
    
    _gameController.controllerPausedHandler = ^(GCController* controller)
    {
        [weakSelf clickPlayPauseButton:nil];
        
        if ([theVICEMachine isPaused])
            [weakSelf controlsFadeIn];
        else
            [weakSelf controlsFadeOut];
    };
}


// ----------------------------------------------------------------------------
- (UIBarPosition) positionForBar:(id<UIBarPositioning>)bar
// ----------------------------------------------------------------------------
{
    if (bar == _toolBar)
        return UIBarPositionTopAttached;
    else
        return UIBarPositionAny;
}


// ----------------------------------------------------------------------------
- (BOOL) shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
// ----------------------------------------------------------------------------
{
	return YES;
}


// ----------------------------------------------------------------------------
- (void) startEmulator
// ----------------------------------------------------------------------------
{
    [[UIApplication sharedApplication] setStatusBarStyle:UIStatusBarStyleDefault animated:YES];
    [UIApplication sharedApplication].idleTimerDisabled = YES;
    
    [self startVICEThread:_supportedFiles];
}


// ----------------------------------------------------------------------------
- (void) startVICEThread:(NSArray*)files
// ----------------------------------------------------------------------------
{
    if (files == nil || files.count == 0)
        return;
    
    _canTerminate = NO;
    
    // Just pick the first one for now, later on, we might do some better guess by looking
    // at the end of the filenames for _0 or _A or something.
    _dataFilePath = [files objectAtIndex:0];
    
    NSString* documentsPath = [NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES) objectAtIndex:0];
    const char* documentsPathCString = [documentsPath cStringUsingEncoding:NSUTF8StringEncoding];
    setenv("HOME", documentsPathCString, 1);
    
    NSString* rootPath = [[[NSBundle mainBundle] resourcePath] stringByAppendingPathComponent:@"x64"];

    if ([[_dataFilePath pathExtension] caseInsensitiveCompare:@"sid"] == NSOrderedSame)
    {
        // disabled, doesn't work properly yet
        [self dismiss];
        return;
        
//        _arguments = [NSArray arrayWithObjects:rootPath, _dataFilePath, nil];
//        c64scene_fake_vsid = 1;
//        _sidInfoView.hidden = NO;
//        _joystickModeButton.enabled = NO;
//        _joystickModeButton2.enabled = NO;
//        _keyboardButton.enabled = NO;
//        _keyboardButton2.enabled = NO;
//        _settingsButton.enabled = NO;
//        _settingsButton2.enabled = NO;
//        _oscilloscopeView.hidden = NO;
//        _oscilloscopeTimer = [NSTimer scheduledTimerWithTimeInterval:1.0f/60.0f target:self selector:@selector(updateOscilloscope) userInfo:nil repeats:YES];
//        [[NSRunLoop currentRunLoop] addTimer:_oscilloscopeTimer forMode:UITrackingRunLoopMode];
    }
    else
    {
        if ([[_dataFilePath pathExtension] caseInsensitiveCompare:@"crt"] == NSOrderedSame)
            _arguments = [NSArray arrayWithObjects:rootPath, @"-cartcrt", _dataFilePath, nil];
        else
            _arguments = [NSArray arrayWithObjects:rootPath, @"-autostart", _dataFilePath, nil];
        c64scene_fake_vsid = 0;
        _sidInfoView.hidden = YES;
        _joystickModeButton.enabled = YES;
        _joystickModeButton2.enabled = YES;
        _keyboardButton.enabled = YES;
        _keyboardButton2.enabled = YES;
        _settingsButton.enabled = YES;
        _settingsButton2.enabled = YES;
        _oscilloscopeView.hidden = YES;
        [_oscilloscopeTimer invalidate];
    }
    
    VICEMachine* viceMachine = [VICEMachine sharedInstance];
    [viceMachine setMediaFiles:files];
    [viceMachine setAutoStartPath:_dataFilePath];
    
    [NSThread detachNewThreadSelector:@selector(startMachine:) toTarget:viceMachine withObject:self];

    _emulationRunning = YES;

    [self scheduleControlFadeOut:0.0f];
}


// ----------------------------------------------------------------------------
- (void) startDataDownload
// ----------------------------------------------------------------------------
{
    NSURL* url = [NSURL URLWithString:_dataFileURLString];
    if (url.isFileURL)
    {
        _supportedFiles = [NSMutableArray arrayWithObject:url.path];
        [_loadingView removeFromSuperview];
        _loadingView = nil;
        
        [self startEmulator];
        return;
    }
        
        
    NSMutableURLRequest* request = [NSMutableURLRequest requestWithURL:url cachePolicy:NSURLRequestUseProtocolCachePolicy timeoutInterval:60.0];
    _downloadData = [NSMutableData data];
    _downloadUrlConnection = [[NSURLConnection alloc] initWithRequest:request delegate:self];
}


// ----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didReceiveResponse:(NSURLResponse*)response
// ----------------------------------------------------------------------------
{
    if (connection == _downloadUrlConnection)
    {
        [_downloadData setLength:0];
        _downloadFilename = [response suggestedFilename];
        return;
    }
}


// ----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didReceiveData:(NSData*)data
// ----------------------------------------------------------------------------
{
    if (connection == _downloadUrlConnection)
    {
        [_downloadData appendData:data];
        return;
    }
    
}


// ----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didFailWithError:(NSError*)error
// ----------------------------------------------------------------------------
{
    //NSLog(@"download connection failed\n");

    NSArray* cachedDownloadFiles = [[CSDatabaseCache sharedInstance] readDownloadFilesForReleaseId:_releaseId];
    if (cachedDownloadFiles != nil)
    {
        NSString* releasePath = [[CSDatabaseCache sharedInstance] releasePathForReleaseId:_releaseId];
        _supportedFiles = [[NSMutableArray alloc] initWithCapacity:cachedDownloadFiles.count];
        
        BOOL all_files_exist = YES;
        
        for (NSString* cachedDownloadFile in cachedDownloadFiles)
        {
            NSString* dataFilePath = [releasePath stringByAppendingPathComponent:cachedDownloadFile];
            BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:dataFilePath isDirectory:NULL];
            if (!exists)
            {
                all_files_exist = NO;
                break;
            }
            [_supportedFiles addObject:dataFilePath];
        }
        
        [_loadingView removeFromSuperview];
        _loadingView = nil;
        
        if (all_files_exist)
            [self startEmulator];
        else
        {
            //NSLog(@"Incomplete downloads!");
            [self dismiss];
        }
    }
    else
        [[CSDatabaseCache sharedInstance] showNetworkError];
}


// ----------------------------------------------------------------------------
- (void) connectionDidFinishLoading:(NSURLConnection*)connection
// ----------------------------------------------------------------------------
{
    if (connection == _downloadUrlConnection && [_downloadData length] > 0)
    {
        //NSLog(@"Request for download finished\n");
        
        NSString* releasePath = [[CSDatabaseCache sharedInstance] releasePathForReleaseId:_releaseId];
        NSString* downloadFilePath = [releasePath stringByAppendingPathComponent:_downloadFilename];
        
        [_downloadData writeToFile:downloadFilePath atomically:YES];
        
        if ([[_downloadFilename pathExtension] caseInsensitiveCompare:@"zip"] == NSOrderedSame)
        {
            NSString* folderName = [_downloadFilename stringByDeletingPathExtension];
            NSString* extractedFolderPath = [releasePath stringByAppendingPathComponent:folderName];
            [[NSFileManager defaultManager] createDirectoryAtPath:extractedFolderPath withIntermediateDirectories:YES attributes:nil error:NULL];
            
            BOOL success = [SSZipArchive unzipFileAtPath:downloadFilePath toDestination:extractedFolderPath];
            if (success)
            {
                _supportedFiles = [[NSMutableArray alloc] init];
                [self findAllSupportedFilesinFolder:extractedFolderPath into:_supportedFiles];
                [[NSFileManager defaultManager] removeItemAtPath:downloadFilePath error:NULL];
            }
        }
        else
            _supportedFiles = [NSMutableArray arrayWithObject:downloadFilePath];
        
        
        [_loadingView removeFromSuperview];
        _loadingView = nil;
        
        if (_supportedFiles.count > 0)
        {
            [[CSDatabaseCache sharedInstance] writeDownloadFiles:_supportedFiles forReleaseId:_releaseId];
            [self startEmulator];
        }
        
        
    }
}



// ----------------------------------------------------------------------------
- (void) findAllSupportedFilesinFolder:(NSString*)extractedFolderPath into:(NSMutableArray*)pathList
// ----------------------------------------------------------------------------
{
    NSArray* supportedExtensions = [NSArray arrayWithObjects:@"d64", @"t64", @"p00", @"prg", @"d71", @"d81", @"sid", @"crt", nil];
    
    NSArray* extractedFiles = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:extractedFolderPath error:NULL];
    extractedFiles = [extractedFiles sortedArrayUsingSelector:@selector(compare:)];
    
    for (NSString* extractedFile in extractedFiles)
    {
        NSString* extractedFilePath = [extractedFolderPath stringByAppendingPathComponent:extractedFile];
        
        BOOL is_directory = NO;
        [[NSFileManager defaultManager] fileExistsAtPath:extractedFilePath isDirectory:&is_directory];
        if (is_directory)
            [self findAllSupportedFilesinFolder:extractedFilePath into:pathList];
        
        NSString* extension = [extractedFile pathExtension];
        
        for (NSString* supportedExtension in supportedExtensions)
        {
            if ([supportedExtension caseInsensitiveCompare:extension] == NSOrderedSame)
            {
                //NSLog(@"Found supported file: %@", extractedFilePath);
                [pathList addObject:extractedFilePath];
                break;
            }
        }
    }
}

extern "C" short* gGetCurrentAudioBuffer();

// ----------------------------------------------------------------------------
- (void) updateOscilloscope
// ----------------------------------------------------------------------------
{
    [_oscilloscopeView setAudioBuffer:_emulationRunning ? gGetCurrentAudioBuffer() : NULL];
    [_oscilloscopeView setNeedsDisplay];
}

// ----------------------------------------------------------------------------
- (BOOL) isRunning
// ----------------------------------------------------------------------------
{
    return _emulationRunning;
}


// ----------------------------------------------------------------------------
- (void) dismiss
// ----------------------------------------------------------------------------
{
    if (![self isBeingDismissed]) {
        dispatch_async(dispatch_get_main_queue(), ^{
            [self dismissViewControllerAnimated:YES completion:^{
            }];
        });
    }
    //[self.presentingViewController dismissViewControllerAnimated:YES completion:NULL];

    if (_releaseViewController != nil)
        _releaseViewController.isPresentingEmulationView = NO;
    [ReleaseViewController attemptRotationToDeviceOrientation];
}


// ----------------------------------------------------------------------------
- (IBAction)clickDoneButton:(id)sender
// ----------------------------------------------------------------------------
{
    if (_settingsPopOver != nil && _settingsPopOver.popoverVisible)
    {
        [_settingsPopOver dismissPopoverAnimated:YES];
        _settingsPopOver = nil;
    }
    
    [_oscilloscopeTimer invalidate];
    [_oscilloscopeView setAudioBuffer:NULL];
    
    if (_canTerminate || !_emulationRunning)
    {
        [self dismiss];
        return;
    }
    
    if (_emulationRunning)
        [theVICEMachine stopMachine];
}


// ----------------------------------------------------------------------------
- (IBAction) clickSettingsButton:(id)sender
// ----------------------------------------------------------------------------
{
    if (_settingsPopOver != nil && _settingsPopOver.popoverVisible)
    {
        [_settingsPopOver dismissPopoverAnimated:YES];
        _settingsPopOver = nil;
    }
    else if (_emulationRunning)
        [self performSegueWithIdentifier:@"SettingsPopOver" sender:sender];
}


// ----------------------------------------------------------------------------
- (void)prepareForSegue:(UIStoryboardSegue *)segue sender:(id)sender
// ----------------------------------------------------------------------------
{
    if (!INTERFACE_IS_PHONE)
    {
        if ([segue.identifier isEqualToString:@"SettingsPopOver"])
            _settingsPopOver = ((UIStoryboardPopoverSegue*)segue).popoverController;
    }
}


// ----------------------------------------------------------------------------
- (IBAction) clickKeyboardButton:(id)sender
// ----------------------------------------------------------------------------
{
    if ([_viceView isFirstResponder])
        [_viceView resignFirstResponder];
    else
        [_viceView becomeFirstResponder];
}



static NSString* sJoystickButtonImageNames[JOYSTICK_MODE_COUNT] = 
{
    @"joystick",
    @"joystick_1",
    @"joystick_2",
};


static uint8_t sJoystickState = 0;


// ----------------------------------------------------------------------------
static inline void sSetJoystickBit(int bit)
// ----------------------------------------------------------------------------
{
    sJoystickState |= bit;
}


// ----------------------------------------------------------------------------
static inline void sClearJoystickBit(int bit)
// ----------------------------------------------------------------------------
{
    sJoystickState &= ~bit;
}


// ----------------------------------------------------------------------------
static inline BOOL sIsJoystickBitSet(int bit)
// ----------------------------------------------------------------------------
{
    return (sJoystickState & bit) == bit;
}


// ----------------------------------------------------------------------------
static inline void sUpdateJoystickState(int port)
// ----------------------------------------------------------------------------
{
    joy_set_bits(port, sJoystickState);
}


// ----------------------------------------------------------------------------
- (void) fadeInVirtualJoystickControls
// ----------------------------------------------------------------------------
{
    if (_joystickMode == JOYSTICK_DISABLED)
        return;
        
    [UIView beginAnimations:@"FadeJoystickIn" context:nil];
    [UIView setAnimationDuration:0.1f];
    [UIView setAnimationCurve:UIViewAnimationCurveLinear];
    _joystickView.alpha = 1.0f;
    _fireButtonView.alpha = 1.0f;
    [UIView commitAnimations];
}


// ----------------------------------------------------------------------------
- (void) fadeOutVirtualJoystickControls
// ----------------------------------------------------------------------------
{
    [UIView beginAnimations:@"FadeJoystickOut" context:nil];
    [UIView setAnimationDuration:0.1f];
    [UIView setAnimationCurve:UIViewAnimationCurveLinear];
    _joystickView.alpha = 0.0f;
    _fireButtonView.alpha = 0.0f;
    [UIView commitAnimations];
}


// ----------------------------------------------------------------------------
- (IBAction) clickJoystickButton:(id)sender
// ----------------------------------------------------------------------------
{
    _joystickMode = (EJoystickMode)((_joystickMode + 1) % JOYSTICK_MODE_COUNT);
    
    _joystickModeButton.image = [UIImage imageNamed:sJoystickButtonImageNames[_joystickMode]];
    _joystickModeButton2.image = [UIImage imageNamed:sJoystickButtonImageNames[_joystickMode]];
    
    if (_joystickMode != JOYSTICK_DISABLED)
    {
        if (_gameController == nil)
            [self fadeInVirtualJoystickControls];
    }
    else
        [self fadeOutVirtualJoystickControls];
}


// ----------------------------------------------------------------------------
- (IBAction) clickPlayPauseButton:(id)sender
// ----------------------------------------------------------------------------
{
    [theVICEMachine togglePause];
    UIImage* image = [theVICEMachine isPaused] ? [UIImage imageNamed:@"hud_play"] : [UIImage imageNamed:@"hud_pause"];
    [_playPauseButton setImage:image forState:UIControlStateNormal];
    [_playPauseButton2 setImage:image forState:UIControlStateNormal];
    [self scheduleControlFadeOut:0.0f];
}


// ----------------------------------------------------------------------------
- (IBAction) clickRestartButton:(id)sender
// ----------------------------------------------------------------------------
{
    [theVICEMachine restart:_dataFilePath];
    [self scheduleControlFadeOut:0.0f];
}



// ----------------------------------------------------------------------------
- (IBAction) clickWarpButton:(id)sender
// ----------------------------------------------------------------------------
{
    _warpEnabled = [theVICEMachine toggleWarpMode];
    UIImage* image = _warpEnabled ? [UIImage imageNamed:@"hud_warp_highlighted"] : [UIImage imageNamed:@"hud_warp"];
    [_warpButton setImage:image forState:UIControlStateNormal];
    [_warpButton2 setImage:image forState:UIControlStateNormal];
    [self scheduleControlFadeOut:0.0f];
}


// ----------------------------------------------------------------------------
- (void) warpStatusNotification:(NSNotification*)notification
// ----------------------------------------------------------------------------
{
    BOOL warp = [[[notification userInfo] objectForKey:@"warp_enabled"] boolValue];
    if (warp != _warpEnabled)
    {
        UIImage* image = warp ? [UIImage imageNamed:@"hud_warp_highlighted"] : [UIImage imageNamed:@"hud_warp"];

        dispatch_async(dispatch_get_main_queue(), ^{
            [_warpButton setImage:image forState:UIControlStateNormal];
            [_warpButton2 setImage:image forState:UIControlStateNormal];
        });

        _warpEnabled = warp;
    }
}


// ----------------------------------------------------------------------------
- (UIStatusBarStyle) preferredStatusBarStyle
// ----------------------------------------------------------------------------
{
    return UIStatusBarStyleLightContent;
}


// ----------------------------------------------------------------------------
- (BOOL) prefersStatusBarHidden
// ----------------------------------------------------------------------------
{
    return !_controlsVisible;
}


// ----------------------------------------------------------------------------
- (BOOL) controlsVisible
// ----------------------------------------------------------------------------
{
    return _controlsVisible;
}


static const float sVisibleAlpha = 1.0f;


// ----------------------------------------------------------------------------
- (void) controlsFadeIn
// ----------------------------------------------------------------------------
{
    CGRect joystickFrame = _joystickView.frame;
    CGRect fireButtonFrame = _fireButtonView.frame;
    
    float ypos = 0.0f;
    float baseToolbarPosition = 0.0f;

    if (UIInterfaceOrientationIsPortrait(self.interfaceOrientation))
        baseToolbarPosition = INTERFACE_IS_PHONE ? _portraitBottomToolBarView.frame.origin.y : _bottomToolBar.frame.origin.y;
    else if (UIInterfaceOrientationIsLandscape(self.interfaceOrientation))
        baseToolbarPosition = _bottomToolBar.frame.origin.y;

    ypos = baseToolbarPosition - (INTERFACE_IS_PHONE ? 168.0f : 180.0f);

    joystickFrame.origin.y = ypos;
    fireButtonFrame.origin.y = ypos;
    
    [UIView beginAnimations:@"FadeIn" context:nil];
    [UIView setAnimationDuration:0.2f];
    [UIView setAnimationCurve:UIViewAnimationCurveLinear];
    _toolBar.alpha = sVisibleAlpha;
    _bottomToolBar.alpha = sVisibleAlpha;
    _portraitBottomToolBarView.alpha = sVisibleAlpha;
    _titleLabel.alpha = sVisibleAlpha;
    _driveLedView.alpha = sVisibleAlpha;
    
    _joystickView.frame = joystickFrame;
    _fireButtonView.frame = fireButtonFrame;

    [UIView commitAnimations];
    
    _controlsVisible = YES;
    
    [self scheduleControlFadeOut:0.0f];
    
    [self setNeedsStatusBarAppearanceUpdate];
}


// ----------------------------------------------------------------------------
- (void) controlsFadeOut
// ----------------------------------------------------------------------------
{
    // Don't fade out if popover is visible
    if (_settingsPopOver != nil && _settingsPopOver.popoverVisible)
    {
        [self scheduleControlFadeOut:0.0f];
        return;
    }
    
    CGRect joystickFrame = _joystickView.frame;
    CGRect fireButtonFrame = _fireButtonView.frame;
    
    float ypos = [self.view bounds].size.height - (INTERFACE_IS_PHONE ? 168.0f : 180.0f);
    
    joystickFrame.origin.y = ypos;
    fireButtonFrame.origin.y = ypos;
    
    [UIView beginAnimations:@"FadeOut" context:nil];
    [UIView setAnimationCurve:UIViewAnimationCurveLinear];
    [UIView setAnimationDuration:0.2f];
    _toolBar.alpha = 0.0f;
    _bottomToolBar.alpha = 0.0f;
    _portraitBottomToolBarView.alpha = 0.0f;
    _titleLabel.alpha = 0.0f;
    _driveLedView.alpha = 0.0f;
    
    _joystickView.frame = joystickFrame;
    _fireButtonView.frame = fireButtonFrame;
    
    [UIView commitAnimations];    
    
    _controlsVisible = NO;
    [self setNeedsStatusBarAppearanceUpdate];
}


// ----------------------------------------------------------------------------
- (void) scheduleControlFadeIn:(float)delay
// ----------------------------------------------------------------------------
{
    // Schedule fade out after 5 seconds
    if (_controlFadeTimer != nil)
        [_controlFadeTimer invalidate]; 
    _controlFadeTimer = [NSTimer scheduledTimerWithTimeInterval:(delay == 0.0f ? sControlsFadeDelay : delay) target:self selector:@selector(controlsFadeIn) userInfo:nil repeats:NO];
}


// ----------------------------------------------------------------------------
- (void) scheduleControlFadeOut:(float)delay
// ----------------------------------------------------------------------------
{
    // Schedule fade out after 5 seconds
    if (_controlFadeTimer != nil)
        [_controlFadeTimer invalidate];
    _controlFadeTimer = [NSTimer scheduledTimerWithTimeInterval:(delay == 0.0f ? sControlsFadeDelay : delay) target:self selector:@selector(controlsFadeOut) userInfo:nil repeats:NO];
}


// ----------------------------------------------------------------------------
- (void) cancelControlFade
// ----------------------------------------------------------------------------
{
    if (_controlFadeTimer != nil)
        [_controlFadeTimer invalidate]; 
}


// ----------------------------------------------------------------------------
- (void)viewWillTransitionToSize:(CGSize)size withTransitionCoordinator:(id<UIViewControllerTransitionCoordinator>)coordinator
// ----------------------------------------------------------------------------
{
    [super viewWillTransitionToSize:size withTransitionCoordinator:coordinator];
    
    // Code here will execute before the rotation begins.
    // Equivalent to placing it in the deprecated method -[willRotateToInterfaceOrientation:duration:]

    if (size.width > size.height)
    {
        [self setLandscapeViceViewFrame:0.1f animCurve:UIViewAnimationCurveLinear canvasSize:_viceView.textureSize];
        if (INTERFACE_IS_PHONE)
        {
            _bottomToolBar.hidden = NO;
            _portraitBottomToolBarView.hidden = YES;
        }
    }
    else if (size.width < size.height)
    {

        [self setPortraitViceViewFrame:0.1f animCurve:UIViewAnimationCurveLinear canvasSize:_viceView.textureSize];
        if (INTERFACE_IS_PHONE)
        {
            _bottomToolBar.hidden = YES;
            _portraitBottomToolBarView.hidden = NO;
        }
    }
    
    [coordinator animateAlongsideTransition:^(id<UIViewControllerTransitionCoordinatorContext> context)
    {
        // Place code here to perform animations during the rotation.
        // You can pass nil or leave this block empty if not necessary.
        
    }
    completion:^(id<UIViewControllerTransitionCoordinatorContext> context)
    {
        // Code here will execute after the rotation has finished.
        // Equivalent to placing it in the deprecated method -[didRotateFromInterfaceOrientation:]
    }];
}



// ----------------------------------------------------------------------------
- (void) shrinkOrGrowViceView
// ----------------------------------------------------------------------------
{
    BOOL allowShrinkGrow = NO;
    if (INTERFACE_IS_PHONE)
    {
        allowShrinkGrow = YES;
    }
    else
    {
        allowShrinkGrow = UIInterfaceOrientationIsLandscape(self.interfaceOrientation);
    }
    
    if (allowShrinkGrow && !_keyboardVisible)
    {
        _viceViewScaled = !_viceViewScaled;

        if (UIInterfaceOrientationIsPortrait(self.interfaceOrientation))
        {
            [self setPortraitViceViewFrame:0.1f animCurve:UIViewAnimationCurveLinear canvasSize:_viceView.textureSize];
        }
        else if (UIInterfaceOrientationIsLandscape(self.interfaceOrientation))
        {
            [self setLandscapeViceViewFrame:0.1f animCurve:UIViewAnimationCurveLinear canvasSize:_viceView.textureSize];
        }
    }
}


// UIDeviceOrientationUnknown                0
// UIDeviceOrientationPortrait               1
// UIDeviceOrientationPortraitUpsideDown     2
// UIDeviceOrientationLandscapeLeft          3
// UIDeviceOrientationLandscapeRight         4


// ----------------------------------------------------------------------------
- (void) setPortraitViceViewFrame:(CGFloat)duration animCurve:(UIViewAnimationCurve)curve canvasSize:(CGSize)canvasSize
// ----------------------------------------------------------------------------
{
    //NSLog(@"setPortrait with orientation: %ld", (long)self.interfaceOrientation);
    
    float superViewWidth = [self.view bounds].size.width;
    float superViewHeight = [self.view bounds].size.height;
    
    CGRect frame = _viceView.frame;
    
    if (INTERFACE_IS_PHONE)
    {
        double aspectRatio = canvasSize.width / canvasSize.height;
        float width = MIN(superViewWidth, superViewHeight);
        
        if (width < canvasSize.width)
            frame.size.width = width;         // on iPhone 4/5 screen is too small to show full res in portrait, so no scaling
        else
            frame.size.width = _viceViewScaled ? MAX(width, canvasSize.width) : MIN(width, canvasSize.width); // on iPhone 6 the screen shows 384 by default and can be scaled to full width

        frame.size.height = frame.size.width / aspectRatio;
        
        float ypos = (superViewHeight - frame.size.height) / 2.0f;
        float bottom = ypos + frame.size.height;
        float keyboardTop = superViewHeight - _keyboardOffset;

        frame.origin.x = (superViewWidth - frame.size.width) / 2.0f;

        if (keyboardTop < bottom)
            frame.origin.y = ypos - (bottom - keyboardTop);
        else
            frame.origin.y = ypos;
    }
    else
    {
        if (_keyboardOffset > 0.0f)
        {
            float ypos = (superViewHeight - frame.size.height) / 2.0f;
            float bottom = ypos + frame.size.height;
            float keyboardTop = superViewHeight - _keyboardOffset;

            frame.size.width = canvasSize.width * 2.0f;
            frame.size.height = canvasSize.height * 2.0f;
            frame.origin.x = (superViewWidth - frame.size.width) / 2.0f;
            
            if (keyboardTop < bottom)
                frame.origin.y = ypos - (bottom - keyboardTop);
            else
                frame.origin.y = ypos;
        }
        else
        {
            frame.size.width = canvasSize.width * 2.0f;
            frame.size.height = canvasSize.height * 2.0f;
            frame.origin.x = (superViewWidth - frame.size.width) / 2.0f;
            frame.origin.y = (superViewHeight - frame.size.height) / 2.0f;
        }
    }

    //NSLog(@"portrait superview: %f/%f, vice frame: %@", superViewWidth, superViewHeight, NSStringFromCGRect(frame));
    
    CGRect joystickFrame = _joystickView.frame;
    CGRect fireButtonFrame = _fireButtonView.frame;
    
    float baseToolbarPosition = INTERFACE_IS_PHONE ? _portraitBottomToolBarView.frame.origin.y : _bottomToolBar.frame.origin.y;
    float ypos = (_controlsVisible ? baseToolbarPosition : [self.view bounds].size.height) - (INTERFACE_IS_PHONE ? 168.0f : 180.0f);
    joystickFrame.origin.y = ypos;
    fireButtonFrame.origin.y = ypos;
    
    [UIView beginAnimations:@"Resize" context:nil];
    [UIView setAnimationDuration:duration];
    [UIView setAnimationCurve:curve];

    _viceView.frame = frame;
    _joystickView.frame = joystickFrame;
    _fireButtonView.frame = fireButtonFrame;
    
    [UIView commitAnimations];
}


// ----------------------------------------------------------------------------
- (void) setLandscapeViceViewFrame:(CGFloat)duration animCurve:(UIViewAnimationCurve)curve canvasSize:(CGSize)canvasSize
// ----------------------------------------------------------------------------
{
//    NSLog(@"setLandscape with orientation: %ld", (long)self.interfaceOrientation);

    float superViewWidth = [self.view bounds].size.width;
    float superViewHeight = [self.view bounds].size.height;

    CGRect frame = _viceView.frame;
    if (canvasSize.width == 0.0f || canvasSize.height == 0.0f)
        canvasSize = CGSizeMake(384.0f, 272.0f);
    
    double aspectRatio = canvasSize.width / canvasSize.height;
    
    if (INTERFACE_IS_PHONE)
    {
        if (_keyboardOffset > 0.0f)
        {
            frame.size.height = superViewHeight - _keyboardOffset;
            frame.size.width = frame.size.height * aspectRatio;
            frame.origin.x = (superViewWidth - frame.size.width) / 2.0f;
            frame.origin.y = 0.0f;
        }
        else
        {
            frame.size.height = _viceViewScaled ? MIN(superViewWidth, superViewHeight) : canvasSize.height;
            frame.size.width = frame.size.height * aspectRatio;
            frame.origin.x = (superViewWidth - frame.size.width) / 2.0f;
            frame.origin.y = (superViewHeight - frame.size.height) / 2.0f;
        }
    }
    else
    {
        if (_keyboardOffset > 0.0f)
        {
            frame.size.height = superViewHeight - _keyboardOffset;
            frame.size.width = frame.size.height * aspectRatio;
            frame.origin.x = (superViewWidth - frame.size.width) / 2.0f;
            frame.origin.y = 0.0f;
        }
        else
        {
            CGSize doubledCanvasSize = CGSizeMake(canvasSize.width * 2.0f, canvasSize.height * 2.0f);
            frame.size.width = _viceViewScaled ? MAX(superViewWidth, superViewHeight) : doubledCanvasSize.width;
            frame.size.height = frame.size.width / aspectRatio;
            frame.origin.x = (superViewWidth - frame.size.width) / 2.0f;
            frame.origin.y = (superViewHeight - frame.size.height) / 2.0f;
        }
    }

    //NSLog(@"landscape superview: %@, vice frame: %@", NSStringFromCGRect([self.view bounds]), NSStringFromCGRect(frame));

    CGRect joystickFrame = _joystickView.frame;
    CGRect fireButtonFrame = _fireButtonView.frame;
    
    float ypos = (_controlsVisible ? _bottomToolBar.frame.origin.y : [self.view bounds].size.height) - (INTERFACE_IS_PHONE ? 168.0f : 180.0f);
    joystickFrame.origin.y = ypos;
    fireButtonFrame.origin.y = ypos;

    [UIView beginAnimations:@"Resize" context:nil];
    [UIView setAnimationDuration:0.1f];
    [UIView setAnimationCurve:curve];

    _viceView.frame = frame;
    _joystickView.frame = joystickFrame;
    _fireButtonView.frame = fireButtonFrame;
    
    [UIView commitAnimations];
    
    //NSLog(@"frame: %@, center: %@", NSStringFromCGRect(frame), NSStringFromCGPoint(_viceView.center));
}


// ----------------------------------------------------------------------------
- (void) registerForNotifications
// ----------------------------------------------------------------------------
{
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(keyboardWillBeShown:)
                                                 name:UIKeyboardWillShowNotification object:nil];
    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(keyboardWillBeHidden:)
                                                 name:UIKeyboardWillHideNotification object:nil];

    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(enableDriveStatus:)
                                                 name:VICEEnableDriveStatusNotification
                                               object:nil];

    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(displayLed:)
                                                 name:VICEDisplayDriveLedNotification
                                               object:nil];

    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(setSidTitle:)
                                                 name:VICESidTitleNotification
                                               object:nil];

    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(setSidAuthor:)
                                                 name:VICESidAuthorNotification
                                               object:nil];
    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(setSidReleased:)
                                                 name:VICESidReleasedNotification
                                               object:nil];
    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(controllerWasConnected:)
                                                 name:GCControllerDidConnectNotification
                                               object:nil];
    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(controllerWasDisconnected:)
                                                 name:GCControllerDidDisconnectNotification
                                               object:nil];
}


// ----------------------------------------------------------------------------
- (void)keyboardWillBeShown:(NSNotification*) notification
// ----------------------------------------------------------------------------
{
    _keyboardVisible = YES;
    NSDictionary* info = [notification userInfo];
    CGSize keyboard_size = [[info objectForKey:UIKeyboardFrameBeginUserInfoKey] CGRectValue].size;
    
    _keyboardOffset = keyboard_size.height;
    
    double duration = [[info objectForKey:UIKeyboardAnimationDurationUserInfoKey] doubleValue];
    UIViewAnimationCurve curve = [[info objectForKey:UIKeyboardAnimationCurveUserInfoKey] integerValue];

    if (UIInterfaceOrientationIsLandscape(self.interfaceOrientation))
        [self setLandscapeViceViewFrame:duration animCurve:curve canvasSize:_viceView.textureSize];
    else if (UIInterfaceOrientationIsPortrait(self.interfaceOrientation))
        [self setPortraitViceViewFrame:duration animCurve:curve canvasSize:_viceView.textureSize];
    
    [self controlsFadeOut];
}


// ----------------------------------------------------------------------------
- (void)keyboardWillBeHidden:(NSNotification*) notification
// ----------------------------------------------------------------------------
{
    _keyboardVisible = NO;
    _keyboardOffset = 0.0f;

    NSDictionary* info = [notification userInfo];
    double duration = [[info objectForKey:UIKeyboardAnimationDurationUserInfoKey] doubleValue];
    UIViewAnimationCurve curve = [[info objectForKey:UIKeyboardAnimationCurveUserInfoKey] integerValue];
    
    if (UIInterfaceOrientationIsLandscape(self.interfaceOrientation))
        [self setLandscapeViceViewFrame:duration animCurve:curve canvasSize:_viceView.textureSize];
    else
        [self setPortraitViceViewFrame:duration animCurve:curve canvasSize:_viceView.textureSize];
}


// ----------------------------------------------------------------------------
- (UILabel*) titleLabel
// ----------------------------------------------------------------------------
{
    return _titleLabel;
}


// ----------------------------------------------------------------------------
- (void)enableDriveStatus:(NSNotification*)notification
// ----------------------------------------------------------------------------
{
    // setup drive views
    NSDictionary * dict = [notification userInfo];
    int enable = [[dict objectForKey:@"enabled_drives"] intValue];

    _driveLedView.hidden = !(enable == 1);
}


// ----------------------------------------------------------------------------
- (void)displayLed:(NSNotification*)notification
// ----------------------------------------------------------------------------
{
    NSDictionary * dict = [notification userInfo];
    int drive = [[dict objectForKey:@"drive"] intValue];
    
    if (drive == 0)
    {
        int pwm = [[dict objectForKey:@"pwm"] intValue];
        
        float strength = pwm/1000.0f;
        
        UIColor* color = [UIColor colorWithRed:(0.4f + strength*0.5f) green:0.0f blue:0.0f alpha:1.0f];
        _driveLedView.image = [UIImage imageWithColor:color];
    }
}


// ----------------------------------------------------------------------------
- (void)setSidTitle:(NSNotification*)notification
// ----------------------------------------------------------------------------
{
    NSDictionary * dict = [notification userInfo];
   _sidTitleLabel.text = [dict objectForKey:@"title"];
}


// ----------------------------------------------------------------------------
- (void)setSidAuthor:(NSNotification*)notification
// ----------------------------------------------------------------------------
{
    NSDictionary * dict = [notification userInfo];
    _sidAuthorLabel.text = [dict objectForKey:@"author"];
}


// ----------------------------------------------------------------------------
- (void)setSidReleased:(NSNotification*)notification
// ----------------------------------------------------------------------------
{
    NSDictionary * dict = [notification userInfo];
    _sidReleasedLabel.text = [dict objectForKey:@"released"];
}



// ----------------------------------------------------------------------------
- (void) setIsInBackground:(BOOL)isInBackground
// ----------------------------------------------------------------------------
{
    [_viceView setIsRenderingAllowed:!isInBackground];
}


#pragma mark VICE Application Protocol implementation


// ----------------------------------------------------------------------------
- (NSArray*) arguments
// ----------------------------------------------------------------------------
{
    return _arguments;
}


// ----------------------------------------------------------------------------
- (VICEGLView*) viceView
// ----------------------------------------------------------------------------
{
    return _viceView;
}


// ----------------------------------------------------------------------------
-(void) setMachine:(id)aMachineObject
// ----------------------------------------------------------------------------
{
    
}


// ----------------------------------------------------------------------------
-(void) machineDidStop
// ----------------------------------------------------------------------------
{
    _canTerminate = YES;
    _emulationRunning = NO;
    
    dispatch_async(dispatch_get_main_queue(), ^{
        [UIApplication sharedApplication].idleTimerDisabled = NO;
    });

    [self dismiss];
}


// ----------------------------------------------------------------------------
-(void) createCanvas:(NSData *)canvasPtr withSize:(CGSize)size
// ----------------------------------------------------------------------------
{
    
}


// ----------------------------------------------------------------------------
-(void) destroyCanvas:(NSData *)canvasPtr
// ----------------------------------------------------------------------------
{
    
}


// ----------------------------------------------------------------------------
-(void) resizeCanvas:(NSData *)canvasPtr withSize:(CGSize)size
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE)
    {
        dispatch_async(dispatch_get_main_queue(), ^{
            if (UIInterfaceOrientationIsPortrait(self.interfaceOrientation))
                [self setPortraitViceViewFrame:0.1f animCurve:UIViewAnimationCurveLinear canvasSize:size];
            else if (UIInterfaceOrientationIsLandscape(self.interfaceOrientation))
                [self setLandscapeViceViewFrame:0.1f animCurve:UIViewAnimationCurveLinear canvasSize:size];
        });
    }
}


// ----------------------------------------------------------------------------
-(void) reconfigureCanvas:(NSData *)canvasPtr
// ----------------------------------------------------------------------------
{
    
}


// ----------------------------------------------------------------------------
-(void) beginLineInputWithPrompt:(NSString *)prompt
// ----------------------------------------------------------------------------
{
    
}


// ----------------------------------------------------------------------------
-(void) endLineInput
// ----------------------------------------------------------------------------
{
    
}


// ----------------------------------------------------------------------------
- (void) postRemoteNotification:(NSArray *)array
// ----------------------------------------------------------------------------
{
    NSString *notificationName = (NSString *)[array objectAtIndex:0];
    NSDictionary *userInfo = (NSDictionary *)[array objectAtIndex:1];
    
    // post notification in our UI thread's default notification center
    [[NSNotificationCenter defaultCenter] postNotificationName:notificationName
                                                        object:self
                                                      userInfo:userInfo];
}


// ----------------------------------------------------------------------------
- (void) runErrorMessage:(NSString *)message
// ----------------------------------------------------------------------------
{
    //NSLog(@"ERROR: %@", message);
}


// ----------------------------------------------------------------------------
- (void) runWarningMessage:(NSString *)message
// ----------------------------------------------------------------------------
{
    //NSLog(@"WARNING: %@", message);
}


// ----------------------------------------------------------------------------
- (int) runCPUJamDialog:(NSString *)message
// ----------------------------------------------------------------------------
{
    [self clickDoneButton:nil];
    
    return 0;
}


// ----------------------------------------------------------------------------
- (BOOL) runExtendImageDialog
// ----------------------------------------------------------------------------
{
    return NO;
}


// ----------------------------------------------------------------------------
- (NSString *) getOpenFileName:(NSString *)title types:(NSArray *)types
// ----------------------------------------------------------------------------
{
    return nil;
}

@end


@implementation EmulationBackgroundView


// ----------------------------------------------------------------------------
- (void) touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
// ----------------------------------------------------------------------------
{
    _keyboardWasVisible = _emulationViewController.keyboardVisible;
    
    if (_emulationViewController.keyboardVisible)
    {
        [_emulationViewController clickKeyboardButton:self];
        return;
    }
    
    for (UITouch* touch in touches)
    {
        if (touch.tapCount == 2)
        {
            [_emulationViewController cancelControlFade];
            [_emulationViewController shrinkOrGrowViceView];
            [_emulationViewController scheduleControlFadeOut:0.0f];
            return;
        }
    }
}


// ----------------------------------------------------------------------------
- (void) touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
// ----------------------------------------------------------------------------
{
    if (_keyboardWasVisible)
        return;

    for (UITouch* touch in touches)
    {
        if (touch.tapCount == 1)
        {
            if (![_emulationViewController controlsVisible])
                [_emulationViewController scheduleControlFadeIn:0.5f];
            else
                [_emulationViewController scheduleControlFadeOut:0.5f];
        }
    }
}


// ----------------------------------------------------------------------------
- (void) touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event
// ----------------------------------------------------------------------------
{
    
}

@end


@implementation JoystickView


static CGRect sUpButtonRect = CGRectMake(0.0f, 0.0f, 180.0f, 64.0f);
static CGRect sDownButtonRect = CGRectMake(0.0f, 116.0f, 180.0f, 64.0f);
static CGRect sLeftButtonRect = CGRectMake(0.0f, 0.0f, 64.0f, 180.0f);
static CGRect sRightButtonRect = CGRectMake(116.0f, 0.0f, 64.0f, 180.0f);


// ----------------------------------------------------------------------------
- (void) awakeFromNib
// ----------------------------------------------------------------------------
{
    [super awakeFromNib];
    
    _upButtonTouches = [NSMutableArray array];
    _downButtonTouches = [NSMutableArray array];
    _leftButtonTouches = [NSMutableArray array];
    _rightButtonTouches = [NSMutableArray array];
}


// ----------------------------------------------------------------------------
- (void) touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
// ----------------------------------------------------------------------------
{
    for (UITouch* touch in touches)
    {
        CGPoint location = [touch locationInView:self];
        if (CGRectContainsPoint(sUpButtonRect, location))
        {
            sSetJoystickBit(1);
            if (![_upButtonTouches containsObject:touch])
                [_upButtonTouches addObject:touch];
        }
        if (CGRectContainsPoint(sDownButtonRect, location))
        {
            sSetJoystickBit(2);
            if (![_downButtonTouches containsObject:touch])
                [_downButtonTouches addObject:touch];
        }
        if (CGRectContainsPoint(sLeftButtonRect, location))
        {
            sSetJoystickBit(4);
            if (![_leftButtonTouches containsObject:touch])
                [_leftButtonTouches addObject:touch];
        }
        if (CGRectContainsPoint(sRightButtonRect, location))
        {
            sSetJoystickBit(8);
            if (![_rightButtonTouches containsObject:touch])
                [_rightButtonTouches addObject:touch];
        }
    }

    sUpdateJoystickState(_emulationViewController.joystickMode - 1);
}        


// ----------------------------------------------------------------------------
- (void) touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
// ----------------------------------------------------------------------------
{
    for (UITouch* touch in touches)
    {
        // Remove touch from whatever list it is in now
        if ([_upButtonTouches containsObject:touch])
        {
            sClearJoystickBit(1);
            [_upButtonTouches removeObject:touch];
        }
        if ([_downButtonTouches containsObject:touch])
        {
            sClearJoystickBit(2);
            [_downButtonTouches removeObject:touch];
        }
        if ([_leftButtonTouches containsObject:touch])
        {
            sClearJoystickBit(4);
            [_leftButtonTouches removeObject:touch];
        }
        if ([_rightButtonTouches containsObject:touch])
        {
            sClearJoystickBit(8);
            [_rightButtonTouches removeObject:touch];
        }

        // Add to new (or the same) list
        CGPoint location = [touch locationInView:self];
        if (CGRectContainsPoint(sUpButtonRect, location))
        {
            sSetJoystickBit(1);
            if (![_upButtonTouches containsObject:touch])
                [_upButtonTouches addObject:touch];
        }
        if (CGRectContainsPoint(sDownButtonRect, location))
        {
            sSetJoystickBit(2);
            if (![_downButtonTouches containsObject:touch])
                [_downButtonTouches addObject:touch];
        }
        if (CGRectContainsPoint(sLeftButtonRect, location))
        {
            sSetJoystickBit(4);
            if (![_leftButtonTouches containsObject:touch])
                [_leftButtonTouches addObject:touch];
        }
        if (CGRectContainsPoint(sRightButtonRect, location))
        {
            sSetJoystickBit(8);
            if (![_rightButtonTouches containsObject:touch])
                [_rightButtonTouches addObject:touch];
        }
    }
    
    sUpdateJoystickState(_emulationViewController.joystickMode - 1);
}


// ----------------------------------------------------------------------------
- (void) touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
// ----------------------------------------------------------------------------
{
    for (UITouch* touch in touches)
    {
        if ([_upButtonTouches containsObject:touch])
        {
            sClearJoystickBit(1);
            [_upButtonTouches removeObject:touch];
        }
        if ([_downButtonTouches containsObject:touch])
        {
            sClearJoystickBit(2);
            [_downButtonTouches removeObject:touch];
        }
        if ([_leftButtonTouches containsObject:touch])
        {
            sClearJoystickBit(4);
            [_leftButtonTouches removeObject:touch];
        }
        if ([_rightButtonTouches containsObject:touch])
        {
            sClearJoystickBit(8);
            [_rightButtonTouches removeObject:touch];
        }
    }
    
    sUpdateJoystickState(_emulationViewController.joystickMode - 1);
}


@end


@implementation FireButtonView


// ----------------------------------------------------------------------------
- (void) awakeFromNib
// ----------------------------------------------------------------------------
{
    [super awakeFromNib];
    
    _fireButtonTouches = [NSMutableArray array];
}


static CGRect sFireButtonRect = CGRectMake(25.0f, 25.0f, 110.0f, 110.0f);


// ----------------------------------------------------------------------------
- (void) touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
// ----------------------------------------------------------------------------
{
    BOOL has_handled_touch = NO;
    
    for (UITouch* touch in touches)
    {
        CGPoint location = [touch locationInView:self];
        if (CGRectContainsPoint(sFireButtonRect, location))
        {
            sSetJoystickBit(16);
            if (![_fireButtonTouches containsObject:touch])
                [_fireButtonTouches addObject:touch];
            has_handled_touch = YES;
            break;
        }
    }
    
    sUpdateJoystickState(_emulationViewController.joystickMode - 1);
    
    if (!has_handled_touch)
        [super touchesBegan:touches withEvent:event];
}        


// ----------------------------------------------------------------------------
- (void) touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
// ----------------------------------------------------------------------------
{
    BOOL has_handled_touch = NO;
    
    for (UITouch* touch in touches)
    {
        if ([_fireButtonTouches containsObject:touch])
        {
            sClearJoystickBit(16);
            [_fireButtonTouches removeObject:touch];
            has_handled_touch = YES;
            break;
        }
    }
    
    sUpdateJoystickState(_emulationViewController.joystickMode - 1);
    
    if (!has_handled_touch)
        [super touchesEnded:touches withEvent:event];
}


@end


@implementation CSVolumeView

// ----------------------------------------------------------------------------
- (CGRect) volumeSliderRectForBounds:(CGRect) bounds
// ----------------------------------------------------------------------------
{
    return UIEdgeInsetsInsetRect(bounds, UIEdgeInsetsMake(0, 0, 0, 40));
}


// ----------------------------------------------------------------------------
- (CGRect) routeButtonRectForBounds:(CGRect) bounds
// ----------------------------------------------------------------------------
{
    CGRect rect = UIEdgeInsetsInsetRect(bounds, UIEdgeInsetsMake(0, bounds.origin.x + bounds.size.width - 40, 0, 0));
    return rect;
}

@end

