//
//  ReleaseViewController.m
//  C64SceneBrowser
//
//  Created by Andreas Varga on 19.03.12.
//  Copyright (c) 2012 __MyCompanyName__. All rights reserved.
//

#import <QuartzCore/CALayer.h>
#import "ReleaseViewController.h"
#import "DetailViewController.h"
#import "MasterViewController.h"
#import "CreditsViewController.h"
#import "EmulationViewController.h"
#import "ScenerViewController.h"
#import "CSReleaseItem.h"
#import "CSCommentItem.h"
#import "CSUserItem.h"
#import "CSScenerItem.h"
#import "CSGroupItem.h"
#import "CSEventItem.h"
#import "CSDatabaseCache.h"
#import "UIImageView+WebCache.h"
#import "SSLoadingView.h"
#import "OHAttributedLabel.h"
#import "AppDelegate.h"
#import "DownloadListViewController.h"


@implementation ReleaseViewController

@synthesize releaseItem = _releaseItem;
@synthesize isPresentingEmulationView = _isPresentingEmulationView;
@synthesize creditsViewController = _creditsViewController;
@synthesize selectedReleaseItem = _selectedReleaseItem;
@synthesize selectedScenerItem = _selectedScenerItem;
@synthesize selectedDownloadIndex = _selectedDownloadIndex;


static const int sNumCommentSections = 6;

static NSString* sCommentSections[sNumCommentSections] = 
{
    @"Summaries",
    @"User Comments",
    @"Production Notes",
    @"Goofs",
    @"Hidden Parts",
    @"Trivia"
};


static const int sNumAllowedReleaseTypes = 21;

static NSString* sAllowedReleaseTypes[sNumAllowedReleaseTypes] =
{
    @"C64 Demo",
    @"C64 One-File Demo",
    @"C64 Intro",
    @"C64 1K Intro",
    @"C64 256b Intro",
    @"C64 4K Intro",
    @"C64 Crack intro",
    @"C64 Basic Demo",
    @"C64 Dentro",
    @"C64 Diskmag",
    @"C64 Fake Demo",
    @"C64 Graphics",
    @"C64 Graphics Collection",
    @"C64 Invitation",
    @"C64 Music",
    @"C64 Music Collection",
    @"C64 Misc.",
    @"C64 Game",
    @"C64 Game Preview",
    @"C64 Crack",
    @"C64 Tool"
};




// ----------------------------------------------------------------------------
- (id) initWithNibName:(NSString *)nibNameOrNil bundle:(NSBundle *)nibBundleOrNil
// ----------------------------------------------------------------------------
{
    self = [super initWithNibName:nibNameOrNil bundle:nibBundleOrNil];
    if (self)
    {
        // Custom initialization
        _previousReleaseItem = nil;
        _zoomedScreenshotView = nil;
        _creditsViewController = nil;
        _creditsButton = nil;
        _actionSheet = nil;
        _isPresentingEmulationView = NO;
    }
    return self;
}



// ----------------------------------------------------------------------------
- (void) viewDidLoad
// ----------------------------------------------------------------------------
{
    [super viewDidLoad];
    
    if (_releaseItem != _previousReleaseItem)
    {
        NSTimeInterval cache_age = 0.0f;
        NSTimeInterval timestamp = [[CSDatabaseCache sharedInstance] readTimeStamp:_releaseItem.identifier ofType:_releaseItem.type];
        NSTimeInterval currentTime = [NSDate timeIntervalSinceReferenceDate];
        cache_age = currentTime - timestamp;
        //NSLog(@"Cached release item is %g seconds old", cache_age);

        if (_releaseItem.cached_depth < 2 || cache_age > 600.0)
        {
            _allImagesLoaded = NO;
            _allImagesRequested = NO;
            _isPresentingEmulationView = NO;
            _creditsButton.enabled = NO;
            
            _commentsLoadingView = [[SSLoadingView alloc] initWithFrame:_commentsTableView.bounds];
            [self.view addSubview:_commentsLoadingView];
            _commentsTableView.hidden = YES;
            
            [_releaseItem cacheAndNotify:self withDepth:2];
        }
        
        [[CSDatabaseCache sharedInstance] writeTimeStampFile:_releaseItem];
        
        _previousReleaseItem = _releaseItem;
        
        UILongPressGestureRecognizer* longPress = [[UILongPressGestureRecognizer alloc] initWithTarget:self action:@selector(longPress:)];
        [_playButton addGestureRecognizer:longPress];
    }
    
    [self updateView];
}


// ----------------------------------------------------------------------------
- (void) viewDidUnload
// ----------------------------------------------------------------------------
{
    _titleLabel = nil;
    _screenshotImageView = nil;
    _subtitleLabel = nil;
    _commentsTableView = nil;
    _ratingLabel = nil;
    _ratingView = nil;
    _ratingOverlayView = nil;
    
    [super viewDidUnload];
}


// ----------------------------------------------------------------------------
- (void) viewWillAppear:(BOOL)animated
// ----------------------------------------------------------------------------
{
    [super viewWillAppear:animated];
    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(allImageDownloadsFinished:)
                                                 name:@"AllImageDownloadsFinished"
                                               object:nil];
  
    UIViewController* masterViewController = [[self.splitViewController.viewControllers objectAtIndex:0] topViewController];
    
    if ([masterViewController class] != [CreditsViewController class])
        [masterViewController performSegueWithIdentifier:@"ShowCreditsDetail" sender:self];
    else
    {
        _creditsViewController = (CreditsViewController*) masterViewController;
        _creditsViewController.releaseItem = _releaseItem;
    }
    
    if (_creditsViewController != nil)
    {
        _creditsViewController.selectedItem = nil;
        [_creditsViewController updateView];
    }
    
    AppDelegate* appDelegate = (AppDelegate*)[[UIApplication sharedApplication] delegate];
    if (appDelegate.masterPopoverController != nil)
        [self addMasterPopoverButton:appDelegate.masterPopoverBarButtonItem];
    
}


// ----------------------------------------------------------------------------
- (void) viewWillDisappear:(BOOL)animated
// ----------------------------------------------------------------------------
{
    [super viewWillDisappear:animated];
    
    [[NSNotificationCenter defaultCenter] removeObserver:self name:@"AllImageDownloadsFinished" object:nil];
}


#if 0
// ----------------------------------------------------------------------------
- (BOOL) shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
// ----------------------------------------------------------------------------
{
    //NSLog(@"shouldAutorotateToInterfaceOrientation: %d (%d)", interfaceOrientation, self.interfaceOrientation);
    
    // Don't auto rotate while we're presenting the emulation view, as otherwise the size of the view is fucked up
    // when we get back...
    if (INTERFACE_IS_PAD && _isPresentingEmulationView)
    {
        // Only allow the current orientation in this case
        if (interfaceOrientation == self.interfaceOrientation)
            return YES;
        else
            return NO;
    }
    
    if (INTERFACE_IS_PHONE)
    {
//        if (interfaceOrientation == UIInterfaceOrientationMaskPortrait)
//            return YES;
//        else
            return NO;
        
    }
    
	return YES;
}
#endif


// ----------------------------------------------------------------------------
- (BOOL) shouldAutorotate
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PAD && _isPresentingEmulationView)
    {
        return NO;
    }
    
    return YES;
}


// ----------------------------------------------------------------------------
- (NSUInteger) supportedInterfaceOrientations
// ----------------------------------------------------------------------------
{
//    if (INTERFACE_IS_PHONE)
//        return UIInterfaceOrientationMaskPortrait;
//    else
        return UIInterfaceOrientationMaskAll;
}


// ----------------------------------------------------------------------------
- (UIInterfaceOrientation)preferredInterfaceOrientationForPresentation
// ----------------------------------------------------------------------------
{
//    if (INTERFACE_IS_PHONE)
//        return UIInterfaceOrientationMaskPortrait;
//    else
        return UIInterfaceOrientationMaskAll;
}



// ----------------------------------------------------------------------------
- (void)longPress:(UILongPressGestureRecognizer*)gesture
// ----------------------------------------------------------------------------
{
    if (gesture.state == UIGestureRecognizerStateBegan)
    {
        if (_downloadListPopOver != nil && _downloadListPopOver.popoverVisible)
        {
            [_downloadListPopOver dismissPopoverAnimated:YES];
            _downloadListPopOver = nil;
        }
        else
            [self performSegueWithIdentifier:@"ShowDownloadList" sender:self];
    }
}


// ----------------------------------------------------------------------------
- (void) updateView
// ----------------------------------------------------------------------------
{
    _titleLabel.text = _releaseItem.name;
    self.navigationItem.title = _releaseItem.name;
    
    if (_releaseItem.releaseType != nil)
    {
        NSString* date = [[CSDatabaseCache sharedInstance] stringForDate:_releaseItem.releaseDate short:NO];
        if (date != nil)
            _subtitleLabel.text = [NSString stringWithFormat:@"%@ | %@", _releaseItem.releaseType, date];
        else
            _subtitleLabel.text = [NSString stringWithFormat:@"%@", _releaseItem.releaseType];
    }
    else
        _subtitleLabel.text = @"";
    
    NSMutableArray* linkRanges = [NSMutableArray array];
    NSMutableArray* linkURLs = [NSMutableArray array];
    NSString* released_text = @"Released by: ";
    BOOL first_group = YES;
    for (CSGroupItem* group in _releaseItem.releaseGroups)
    {
        if (group.name == nil)
            continue;
        
        if (first_group)
        {
            released_text = [released_text stringByAppendingString:group.name];
            first_group = NO;
        }
        else
            released_text = [released_text stringByAppendingFormat:@", %@", group.name];
        
        if (group.website != nil)
        {
            NSRange linkRange = NSMakeRange(released_text.length - group.name.length, group.name.length);
            [linkRanges addObject:[NSValue valueWithRange:linkRange]];
            [linkURLs addObject:[NSURL URLWithString:group.website]];
        }
    }
    
    BOOL first_scener = YES;
    if (_releaseItem.releaseGroups.count == 0 && _releaseItem.releaseSceners.count > 0)
    {
        for (CSScenerItem* scener in _releaseItem.releaseSceners)
        {
            if (scener.handle == nil)
                continue;
            
            if (first_scener)
            {
                released_text = [released_text stringByAppendingString:scener.handle];
                first_scener = NO;
            }
            else
                released_text = [released_text stringByAppendingFormat:@", %@", scener.handle];
        }
    }
    
    _groupsLabel.text = released_text;
    _groupsLabel.linkColor = [UIColor colorWithWhite:0.7f alpha:1.0f];
    _groupsLabel.highlightedLinkColor = [UIColor colorWithWhite:0.9f alpha:1.0f];
    int index = 0;
    for (NSValue* linkValue in linkRanges)
    {
        NSURL* url = [linkURLs objectAtIndex:index];
        NSRange linkRange = [linkValue rangeValue];
        [_groupsLabel addCustomLink:url inRange:linkRange];
        
        index++;
    }
    
    if (_releaseItem.rating == nil || [_releaseItem.rating isEqualToString:@""])
    {
        _ratingLabel.text = @"No rating yet";
        _ratingLabel.textAlignment = NSTextAlignmentLeft;
        _ratingView.hidden = YES;
        _ratingOverlayView.hidden = YES;
    }
    else
    {
        float rating = [_releaseItem.rating floatValue];
        _ratingLabel.text = [NSString stringWithFormat:@"%.1f", rating];
        _ratingLabel.textAlignment = NSTextAlignmentRight;
        _ratingView.hidden = NO;
        _ratingOverlayView.hidden = NO;
        
        _ratingView.backgroundColor = [UIColor colorWithPatternImage:[UIImage imageNamed:@"rating_star_grey"]];
        
        CGRect overlay_frame = _ratingView.frame;
        overlay_frame.size.width = _ratingView.frame.size.width * rating / 10.0f;
        _ratingOverlayView.frame = overlay_frame;
        _ratingOverlayView.autoresizingMask = _ratingView.autoresizingMask;
        _ratingOverlayView.backgroundColor = [UIColor colorWithPatternImage:[UIImage imageNamed:@"rating_star"]];
    }
    
    CSEventItem* event = _releaseItem.releasedAtEvent;
    if (event != nil)
        _releasedAtLabel.text = [NSString stringWithFormat:@"Released at: %@", event.name];
    else
        _releasedAtLabel.text = @"";

    _releasedAtLabel.linkColor = [UIColor colorWithWhite:0.7f alpha:1.0f];
    _releasedAtLabel.highlightedLinkColor = [UIColor colorWithWhite:0.9f alpha:1.0f];

    if (_releaseItem.achievement != nil)
        _achievementLabel.text = [NSString stringWithFormat:@"Achieved: %@", _releaseItem.achievement];
    else
        _achievementLabel.text = @"";
    
    for (UIGestureRecognizer* recognizer in _screenshotImageView.gestureRecognizers)
        [_screenshotImageView removeGestureRecognizer:recognizer];
    
    if (_releaseItem.screenshot != nil && _releaseItem.screenshot.length > 0)
    {
        [_screenshotImageView setImageWithURL:[NSURL URLWithString:_releaseItem.screenshot] placeholderImage:nil];
        [_screenshotImageView addGestureRecognizer:[[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(zoomImage:)]];
    }
    else
        _screenshotImageView.image = nil;
    
    
    _allowDownload = NO;
    for (int i=0; i<sNumAllowedReleaseTypes; i++)
    {
        if ([_releaseItem.releaseType caseInsensitiveCompare:sAllowedReleaseTypes[i]] == NSOrderedSame)
        {
            _allowDownload = YES;
            break;
        }
    }
    
    // Explicitly prevent downloading of HVSC
    NSRange hvsc_range = [_releaseItem.name rangeOfString:@"High Voltage Sid Collection" options:NSCaseInsensitiveSearch];
    if (hvsc_range.location != NSNotFound)
        _allowDownload = NO;
    
    _selectedDownloadIndex = NSNotFound;
    BOOL has_downloads = _releaseItem.downloadURLs.count > 0;
    if (has_downloads && _allowDownload)
    {
        NSInteger oldestUploadId = NSIntegerMax;
        NSInteger oldestUploadIndex = NSNotFound;
        NSInteger firstSupportedFileIndex = NSNotFound;
        
//        NSArray* supportedExtensions = [NSArray arrayWithObjects:@"zip", @"d64", @"prg", @"p00", @"t64", @"sid", @"crt", nil];
        NSArray* supportedExtensions = [NSArray arrayWithObjects:@"zip", @"d64", @"prg", @"p00", @"t64", @"crt", nil];
        NSUInteger index = 0;
        for (NSString* urlString in _releaseItem.downloadURLs)
        {
            BOOL supported = [supportedExtensions containsObject:[[urlString pathExtension] lowercaseString]];
            BOOL isIntro = [urlString containsString:@"intros.c64.org"];
            if (supported || isIntro)
            {
                if (firstSupportedFileIndex == NSNotFound)
                    firstSupportedFileIndex = index;
                
                // If name of file is an exact match with release title, then take that file if it's supported
                NSString* baseName = [[urlString lastPathComponent] stringByDeletingPathExtension];
                baseName = [baseName stringByReplacingOccurrencesOfString:@"_" withString:@" "];
                if ([baseName caseInsensitiveCompare:_releaseItem.name] == NSOrderedSame)
                {
                    _selectedDownloadIndex = index;
                    break;
                }

                // Otherwise take the first download that was added
                NSRange internalFileRange = [urlString rangeOfString:@"getinternalfile.php"];
                if (internalFileRange.location != NSNotFound)
                {
                    NSArray* pathComponents = [urlString pathComponents];
                    NSString* uploadIdString = [pathComponents objectAtIndex:pathComponents.count-2];
                    //NSLog(@"uploadId: %@", uploadIdString);
                    NSInteger uploadId = [uploadIdString integerValue];
                    if (uploadId < oldestUploadId)
                    {
                        oldestUploadId = uploadId;
                        oldestUploadIndex = index;
                    }
                }
            }

            index++;
        }
        
        // If we don't have an exact match, use the oldest internal file upload
        if (_selectedDownloadIndex == NSNotFound)
            _selectedDownloadIndex = oldestUploadIndex;
        
        // If we don't have an internal file (in this case we can't determine the oldest link)
        // then just take the first supported file we found
        if (_selectedDownloadIndex == NSNotFound)
            _selectedDownloadIndex = firstSupportedFileIndex;
        
        if (_selectedDownloadIndex != NSNotFound)
            _playButton.hidden = NO;
        else
            _playButton.hidden = YES;
    }
    else
        _playButton.hidden = YES;
    
    [_commentsTableView reloadData];
    if (_creditsViewController != nil)
        [_creditsViewController updateView];
    
    if (INTERFACE_IS_PHONE)
    {
        if (_scrollView != nil)
        {
            _scrollView.contentSize = CGSizeMake(320.0f, _commentsTableView.contentSize.height);
            _commentsTableView.frame.size = CGSizeMake(_commentsTableView.contentSize.width, _commentsTableView.contentSize.height);
            UIView* view = [_scrollView.subviews objectAtIndex:0];
            view.frame.size = CGSizeMake(_commentsTableView.contentSize.width, _commentsTableView.contentSize.height);
        }
    }
}


// ----------------------------------------------------------------------------
- (IBAction) playRelease:(id)sender
// ----------------------------------------------------------------------------
{
    UIStoryboard* storyBoard = [UIStoryboard storyboardWithName:INTERFACE_IS_PHONE ? @"iPhoneStoryboard" : @"MainStoryboard" bundle:nil];

    _emulationViewController = [storyBoard instantiateViewControllerWithIdentifier:@"EmulationViewController"];

    NSString* releaseDownloadUrl = [_releaseItem.downloadURLs objectAtIndex:_selectedDownloadIndex];
//    int i = 0;
//    for (NSString* url in _releaseItem.downloadURLs)
//    {
//        NSLog(@"Download %d: %@", i, url);
//        i++;
//    }
//    NSLog(@"Selected download: %@", releaseDownloadUrl);
    
    _emulationViewController.dataFileURLString = releaseDownloadUrl;
    _emulationViewController.title = _releaseItem.name;
    _emulationViewController.releaseId = _releaseItem.identifier;
    _emulationViewController.releaseViewController = self;
    
    _isPresentingEmulationView = YES;
    
    if (INTERFACE_IS_PHONE)
        [self presentViewController:_emulationViewController animated:YES completion:NULL];
    else
        [self.splitViewController presentViewController:_emulationViewController animated:YES completion:NULL];
}


/*
// ----------------------------------------------------------------------------
- (IBAction) addComment:(id)sender
// ----------------------------------------------------------------------------
{
    if (_addCommentActionSheet != nil)
        return;
    
    _addCommentActionSheet = [[UIActionSheet alloc] initWithTitle:@"Add Comment"
                                                         delegate:self
                                                cancelButtonTitle:@"Cancel"
                                           destructiveButtonTitle:nil
                                                otherButtonTitles:@"User Comment", @"Summary", @"Trivia", @"Production Note", @"Goof", @"Hidden Part", nil];
    
    _addCommentActionSheet.actionSheetStyle = UIActionSheetStyleBlackOpaque;
    [_addCommentActionSheet showFromBarButtonItem:sender animated:YES];
}


// ----------------------------------------------------------------------------
- (void) actionSheet:(UIActionSheet *)actionSheet didDismissWithButtonIndex:(NSInteger)buttonIndex
// ----------------------------------------------------------------------------
{
    NSLog(@"Action sheet dismissed!\n");
    _addCommentActionSheet = nil;
}
*/


// ----------------------------------------------------------------------------
- (IBAction) tapRefreshButton:(id)sender
// ----------------------------------------------------------------------------
{
    if (_releaseItem.cached_depth == 2)
    {
        _allImagesLoaded = NO;
        _allImagesRequested = NO;
        _isPresentingEmulationView = NO;
        _creditsButton.enabled = NO;
        
        _commentsLoadingView = [[SSLoadingView alloc] initWithFrame:_commentsTableView.bounds];
        [self.view addSubview:_commentsLoadingView];
        _commentsTableView.hidden = YES;
        
        [_releaseItem cacheAndNotify:self withDepth:2];

    }
    
    [[CSDatabaseCache sharedInstance] writeTimeStampFile:_releaseItem];
    
    UIViewController* masterViewController = [[self.splitViewController.viewControllers objectAtIndex:0] topViewController];
    
    if ([masterViewController class] != [CreditsViewController class])
        [masterViewController performSegueWithIdentifier:@"ShowCreditsDetail" sender:self];
    else
    {
        _creditsViewController = (CreditsViewController*) masterViewController;
        _creditsViewController.releaseItem = _releaseItem;
    }
    
    if (_creditsViewController != nil)
    {
        _creditsViewController.selectedItem = nil;
        [_creditsViewController updateView];
    }
}


// ----------------------------------------------------------------------------
- (IBAction) tapActionButton:(id)sender
// ----------------------------------------------------------------------------
{
    if (_actionSheet != nil)
    {
        [_actionSheet dismissWithClickedButtonIndex:_actionSheet.cancelButtonIndex animated:YES];
        _actionSheet = nil;
        return;
    }
    
    _actionSheet = [[UIActionSheet alloc] initWithTitle:nil
                                               delegate:self
                                      cancelButtonTitle:nil
                                 destructiveButtonTitle:nil
                                      otherButtonTitles:@"Add to Favorites", @"Open in Safari", nil];
    
    _actionSheet.actionSheetStyle = UIActionSheetStyleBlackOpaque;
    [_actionSheet showFromBarButtonItem:sender animated:YES];
}


// ----------------------------------------------------------------------------
- (void) actionSheet:(UIActionSheet *)actionSheet didDismissWithButtonIndex:(NSInteger)buttonIndex
// ----------------------------------------------------------------------------
{
    //NSLog(@"Action sheet dismissed! button: %ld\n", (long)buttonIndex);
    switch (buttonIndex) 
    {
        case 0:
            [self addToFavorites:self];
            break;
        case 1:
            [self openInSafari:self];
        default:
            break;
    }
    _actionSheet = nil;
}


// ----------------------------------------------------------------------------
- (IBAction) addToFavorites:(id)sender
// ----------------------------------------------------------------------------
{
    [[CSDatabaseCache sharedInstance] addFavorite:_releaseItem.identifier];
}


// ----------------------------------------------------------------------------
- (IBAction) openInSafari:(id)sender
// ----------------------------------------------------------------------------
{
    NSString* release_url_string = [NSString stringWithFormat:@"http://csdb.dk/release/?id=%@", _releaseItem.identifier];
    [[UIApplication sharedApplication] openURL:[NSURL URLWithString:release_url_string]];
}


// ----------------------------------------------------------------------------
- (void) zoomImage:(id)sender
// ----------------------------------------------------------------------------
{
    UIImageView* imageView = (UIImageView*)[sender view];

    // We really! want this window over the top of everything, inc. the nav bar.
    UIWindow* mainWindow = [[UIApplication sharedApplication] keyWindow];
    if (mainWindow == nil)
        mainWindow = [[UIApplication sharedApplication].windows objectAtIndex:0];
    
    UIViewController* rootViewController = mainWindow.rootViewController;
    
    _origScreenshotView = imageView;

    self.navigationController.navigationBar.userInteractionEnabled = NO;
    _commentsTableView.userInteractionEnabled = NO;
    _origScreenshotView.userInteractionEnabled = NO;
    
    CGRect screen_bounds = rootViewController.view.bounds;

#if 0
    CGRect origScreenshotBounds = [[_origScreenshotView superview] convertRect:_origScreenshotView.frame toView:rootViewController.view];
    
    _zoomedScreenshotView = [[UIImageView alloc] initWithFrame:origScreenshotBounds];
    _zoomedScreenshotView.image = _origScreenshotView.image;
    _zoomedScreenshotView.backgroundColor = [UIColor blackColor];
    _zoomedScreenshotView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
    _zoomedScreenshotView.animationImages = _origScreenshotView.animationImages;
    _zoomedScreenshotView.animationDuration = _origScreenshotView.animationDuration;
    _zoomedScreenshotView.animationRepeatCount = _origScreenshotView.animationRepeatCount;
    if (_zoomedScreenshotView.animationImages.count > 0)
        [_zoomedScreenshotView startAnimating];
  
    _zoomedScreenshotView.contentMode = _origScreenshotView.contentMode;
    [[_zoomedScreenshotView layer] setMagnificationFilter:kCAFilterNearest];
    
    [rootViewController.view addSubview:_zoomedScreenshotView];
    
    [UIView beginAnimations:@"Expand" context:nil];
    [UIView setAnimationDuration:0.4f];
    
        [_zoomedScreenshotView setFrame:screen_bounds];
        _zoomedScreenshotView.transform = CGAffineTransformMakeScale(1.0f, 1.0f);

    [UIView setAnimationDelegate:self];
    [UIView setAnimationDidStopSelector:@selector(animationDidStop:finished:context:)];
    [UIView commitAnimations];

#else
    CGRect origScreenshotBounds = [[_origScreenshotView superview] convertRect:_origScreenshotView.frame toView:rootViewController.view];
    _imageScrollView = [[UIScrollView alloc] initWithFrame:origScreenshotBounds];
    
    _zoomedScreenshotView = [[UIImageView alloc] initWithFrame:[[_origScreenshotView superview] convertRect:_origScreenshotView.frame toView:_imageScrollView]];
    _zoomedScreenshotView.image = _origScreenshotView.image;
    _zoomedScreenshotView.backgroundColor = [UIColor blackColor];
    _zoomedScreenshotView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
    _zoomedScreenshotView.contentMode = _origScreenshotView.contentMode;
    _zoomedScreenshotView.animationImages = _origScreenshotView.animationImages;
    _zoomedScreenshotView.animationDuration = _origScreenshotView.animationDuration;
    _zoomedScreenshotView.animationRepeatCount = _origScreenshotView.animationRepeatCount;
    if (_zoomedScreenshotView.animationImages.count > 0)
        [_zoomedScreenshotView startAnimating];
    [[_zoomedScreenshotView layer] setMagnificationFilter:kCAFilterNearest];

    _imageScrollView = [[UIScrollView alloc] initWithFrame:origScreenshotBounds];
    [_imageScrollView addSubview:_zoomedScreenshotView];
    _imageScrollView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
    _imageScrollView.backgroundColor = [UIColor blackColor];
    _imageScrollView.contentSize = _zoomedScreenshotView.frame.size;
    _imageScrollView.minimumZoomScale = 1.0f;
    _imageScrollView.maximumZoomScale = 6.0f;
    _imageScrollView.delegate = self;
    
    [rootViewController.view addSubview:_imageScrollView];
    
    [UIView beginAnimations:@"Expand" context:nil];
    [UIView setAnimationDuration:0.4f];

        [_imageScrollView setFrame:screen_bounds];
        _imageScrollView.transform = CGAffineTransformMakeScale(1.0f, 1.0f);
    
    [UIView setAnimationDelegate:self];
    [UIView setAnimationDidStopSelector:@selector(animationDidStop:finished:context:)];
    [UIView commitAnimations];
#endif
}


// ----------------------------------------------------------------------------
- (void) shrinkImage:(id) sender
// ----------------------------------------------------------------------------
{
    [_imageScrollView removeGestureRecognizer:sender];

    // We really! want this window over the top of everything, inc. the nav bar.
    UIWindow* mainWindow = [[UIApplication sharedApplication] keyWindow];
    if (mainWindow == nil)
        mainWindow = [[UIApplication sharedApplication].windows objectAtIndex:0];
    
    UIViewController* rootViewController = mainWindow.rootViewController;
    
    CGRect origScreenshotBounds = [[_origScreenshotView superview] convertRect:_origScreenshotView.frame toView:rootViewController.view];
    [_imageScrollView setZoomScale:1.0f animated:YES];
    
    [UIView beginAnimations:@"Shrink" context:nil];
    [UIView setAnimationDuration:0.4f];
    
        [_imageScrollView setFrame:origScreenshotBounds];
    
    [UIView setAnimationDelegate:self];
    [UIView setAnimationDidStopSelector:@selector(animationDidStop:finished:context:)];
    [UIView commitAnimations];
    
    self.navigationController.navigationBar.userInteractionEnabled = YES;
    _commentsTableView.userInteractionEnabled = YES;
    _origScreenshotView.userInteractionEnabled = YES;
}


// ----------------------------------------------------------------------------
- (void)animationDidStop:(NSString *)animationID finished:(BOOL)finished context:(void *)context
// ----------------------------------------------------------------------------
{
    if ([animationID isEqual: @"Expand"])
    {
        _imageScrollView.userInteractionEnabled = YES;
        [_imageScrollView addGestureRecognizer:[[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(shrinkImage:)]];
    }
	else if ([animationID isEqual: @"Shrink"])
    {
        [_zoomedScreenshotView removeFromSuperview];
        _zoomedScreenshotView = nil;
        
		[_imageScrollView removeFromSuperview];
        _imageScrollView = nil;
        
        
        [_origScreenshotView addGestureRecognizer:[[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(zoomImage:)]];
        _origScreenshotView = nil;
    }
}


#pragma mark - CSDatabaseItemDelegate protocol


// ----------------------------------------------------------------------------
- (void) databaseItemIsCached:(CSDatabaseItem*)item
// ----------------------------------------------------------------------------
{
    //NSLog(@"Depth 2 finished loading! Reload comments and credits table!\n");
    
    // At this point, all image downloads have been submitted
    _allImagesRequested = YES;
    
    // Check if they're already done
    SDWebImageManager* manager = [SDWebImageManager sharedManager];
    if (![manager hasPendingDownloads])
    {
        // All images are already downloaded (that was quick), make sure we update the views
        [_commentsTableView reloadData];
        if (_creditsViewController != nil)
            [_creditsViewController updateView];
        
        _allImagesLoaded = YES;
    }
    else
    {
        // Expecting more downloads to come in
        _allImagesLoaded = NO;
    }
    
    _creditsButton.enabled = YES;
    
    [self updateView];
   
    [_commentsLoadingView removeFromSuperview];
    _commentsTableView.hidden = NO;
}


// ----------------------------------------------------------------------------
- (void) cachingFailed
// ----------------------------------------------------------------------------
{
    [_commentsLoadingView removeFromSuperview];
    _commentsTableView.hidden = NO;
    
    [[CSDatabaseCache sharedInstance] showNetworkError];
}

#pragma mark -

// ----------------------------------------------------------------------------
- (void) allImageDownloadsFinished:(NSNotification*)notification
// ----------------------------------------------------------------------------
{
    //NSLog(@"all image downloads finished! Reload comments and credits table!\n");
    
    if (!_allImagesRequested)
    {
        //NSLog(@"But not all download submitted yet!");
        return;
    }
    
    if (!_allImagesLoaded)
    {
        _allImagesLoaded = YES;
        
        [_commentsTableView reloadData];
        if (_creditsViewController != nil)
            [_creditsViewController updateView];
        
        _allImagesRequested = NO;
    }
//    else
//        NSAssert(false, @"Image download finished late!");
}


// ----------------------------------------------------------------------------
- (void) addMasterPopoverButton:(UIBarButtonItem*)barButtonItem
// ----------------------------------------------------------------------------
{
    if (self.toolbarItems.count > 0 && [self.toolbarItems objectAtIndex:0] == barButtonItem)
    {
        barButtonItem.title = @"Credits";
        return;
    }
    
    barButtonItem.title = @"Credits";
    NSMutableArray* toolbarItems = [NSMutableArray arrayWithObject:barButtonItem];
    [self setToolbarItems:[toolbarItems arrayByAddingObjectsFromArray:self.toolbarItems] animated:YES];
}


// ----------------------------------------------------------------------------
- (void) removeMasterPopoverButton:(UIBarButtonItem*)barButtonItem
// ----------------------------------------------------------------------------
{
    if (self.toolbarItems.count > 0 && [self.toolbarItems objectAtIndex:0] == barButtonItem)
        [self setToolbarItems:[self.toolbarItems subarrayWithRange:NSMakeRange(1, self.toolbarItems.count-1)] animated:YES];
}


// ----------------------------------------------------------------------------
- (NSMutableArray*) commentsArrayForSection:(NSUInteger)section
// ----------------------------------------------------------------------------
{
    switch (section)
    {
        case 0:
            return _releaseItem.summaries;
        case 1:
            return _releaseItem.comments;
        case 2:
            return _releaseItem.productionNotes;
        case 3:
            return _releaseItem.goofs;
        case 4:
            return _releaseItem.hiddenParts;
        case 5:
            return _releaseItem.trivias;
    }

    return nil;
}



// ----------------------------------------------------------------------------
- (void)prepareForSegue:(UIStoryboardSegue *)segue sender:(id)sender
// ----------------------------------------------------------------------------
{
    if ([[segue identifier] isEqualToString:@"ShowReleaseDetail"])
    {
        ReleaseViewController* releaseViewController = [segue destinationViewController];
        releaseViewController.releaseItem = _selectedReleaseItem;
        
        if (self.toolbarItems.count == 1)
        {
            NSMutableArray* toolbarItems = [NSMutableArray arrayWithObject:[self.toolbarItems objectAtIndex:0]];
            releaseViewController.toolbarItems = [toolbarItems arrayByAddingObjectsFromArray:releaseViewController.toolbarItems];
        }
    }
    else if ([[segue identifier] isEqualToString:@"ShowScenerDetail"])
    {
        ScenerViewController* scenerViewController = [segue destinationViewController];
        scenerViewController.scenerItem = _selectedScenerItem;
    }
    else if ([segue.identifier isEqualToString:@"ShowDownloadList"])
    {
        if (INTERFACE_IS_PHONE)
        {
            _downloadListPopOver = nil;
            UINavigationController* downloadNavController = [segue destinationViewController];
            DownloadListViewController* downloadListViewController = (DownloadListViewController*) downloadNavController.topViewController;
            downloadListViewController.downloadList = _releaseItem.downloadURLs;
            downloadListViewController.downloadListPopOver = nil;
            downloadListViewController.releaseViewController = self;
        }
        else
        {
            _downloadListPopOver = ((UIStoryboardPopoverSegue*)segue).popoverController;
            UINavigationController* downloadNavController = [segue destinationViewController];
            DownloadListViewController* downloadListViewController = (DownloadListViewController*) downloadNavController.topViewController;
            downloadListViewController.downloadList = _releaseItem.downloadURLs;
            downloadListViewController.downloadListPopOver = _downloadListPopOver;
            downloadListViewController.releaseViewController = self;
        }
    }
    else if (INTERFACE_IS_PHONE && [[segue identifier] isEqualToString:@"ShowCreditsDetail"])
    {
        CreditsViewController* creditsViewController = [segue destinationViewController];
        creditsViewController.releaseItem = _releaseItem;
    }
}


#pragma mark TableView delegate and datasource

// ----------------------------------------------------------------------------
- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE)
        return sNumCommentSections + 1;
    else
        return sNumCommentSections;
}


// ----------------------------------------------------------------------------
- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE)
    {
        if (section == 0)
            return 1;
        
        NSMutableArray* comments = [self commentsArrayForSection:section-1];
        return comments != nil ? comments.count : 0;
    }
    else
    {
        NSMutableArray* comments = [self commentsArrayForSection:section];
        return comments != nil ? comments.count : 0;
    }
}


// ----------------------------------------------------------------------------
- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE && indexPath.section == 0)
    {
        MainInfoItemCell* cell = [_commentsTableView dequeueReusableCellWithIdentifier:@"MainInfoItemCell"];
        
        cell.titleLabel.text = _releaseItem.name;
        
        if (_releaseItem.releaseType != nil)
        {
            NSString* date = [[CSDatabaseCache sharedInstance] stringForDate:_releaseItem.releaseDate short:NO];
            if (date != nil)
                cell.subtitleLabel.text = [NSString stringWithFormat:@"%@ | %@", _releaseItem.releaseType, date];
            else
                cell.subtitleLabel.text = [NSString stringWithFormat:@"%@", _releaseItem.releaseType];
        }
        else
            cell.subtitleLabel.text = @"";
        
        NSMutableArray* linkRanges = [NSMutableArray array];
        NSMutableArray* linkURLs = [NSMutableArray array];
        NSString* released_text = @"Released by: ";
        BOOL first_group = YES;
        for (CSGroupItem* group in _releaseItem.releaseGroups)
        {
            if (group.name == nil)
                continue;
            
            if (first_group)
            {
                released_text = [released_text stringByAppendingString:group.name];
                first_group = NO;
            }
            else
                released_text = [released_text stringByAppendingFormat:@", %@", group.name];
            
            if (group.website != nil)
            {
                NSRange linkRange = NSMakeRange(released_text.length - group.name.length, group.name.length);
                [linkRanges addObject:[NSValue valueWithRange:linkRange]];
                [linkURLs addObject:[NSURL URLWithString:group.website]];
            }
        }
        
        BOOL first_scener = YES;
        if (_releaseItem.releaseGroups.count == 0 && _releaseItem.releaseSceners.count > 0)
        {
            for (CSScenerItem* scener in _releaseItem.releaseSceners)
            {
                if (scener.handle == nil)
                    continue;
                
                if (first_scener)
                {
                    released_text = [released_text stringByAppendingString:scener.handle];
                    first_scener = NO;
                }
                else
                    released_text = [released_text stringByAppendingFormat:@", %@", scener.handle];
            }
        }
        
        cell.groupsLabel.text = released_text;
        cell.groupsLabel.linkColor = [UIColor colorWithWhite:0.7f alpha:1.0f];
        cell.groupsLabel.highlightedLinkColor = [UIColor colorWithWhite:0.9f alpha:1.0f];
        int index = 0;
        for (NSValue* linkValue in linkRanges)
        {
            NSURL* url = [linkURLs objectAtIndex:index];
            NSRange linkRange = [linkValue rangeValue];
            [cell.groupsLabel addCustomLink:url inRange:linkRange];
            
            index++;
        }
        
        if (_releaseItem.rating == nil || [_releaseItem.rating isEqualToString:@""])
        {
            cell.ratingLabel.text = @"No rating yet";
            cell.ratingLabel.textAlignment = NSTextAlignmentLeft;
            cell.ratingView.hidden = YES;
            cell.ratingOverlayView.hidden = YES;
        }
        else
        {
            float rating = [_releaseItem.rating floatValue];
            cell.ratingLabel.text = [NSString stringWithFormat:@"%.1f", rating];
            cell.ratingLabel.textAlignment = NSTextAlignmentRight;
            cell.ratingView.hidden = NO;
            cell.ratingOverlayView.hidden = NO;
            
            cell.ratingView.backgroundColor = [UIColor colorWithPatternImage:[UIImage imageNamed:@"rating_star_grey"]];
            
            CGRect overlay_frame = cell.ratingView.frame;
            overlay_frame.size.width = cell.ratingView.frame.size.width * rating / 10.0f;
            cell.ratingOverlayView.frame = overlay_frame;
            cell.ratingOverlayView.autoresizingMask = cell.ratingView.autoresizingMask;
            cell.ratingOverlayView.backgroundColor = [UIColor colorWithPatternImage:[UIImage imageNamed:@"rating_star"]];
        }
        
        CSEventItem* event = _releaseItem.releasedAtEvent;
        if (event != nil)
            cell.releasedAtLabel.text = [NSString stringWithFormat:@"Released at: %@", event.name];
        else
            cell.releasedAtLabel.text = @"";
        
        cell.releasedAtLabel.linkColor = [UIColor colorWithWhite:0.7f alpha:1.0f];
        cell.releasedAtLabel.highlightedLinkColor = [UIColor colorWithWhite:0.9f alpha:1.0f];
        
        if (_releaseItem.achievement != nil)
            cell.achievementLabel.text = [NSString stringWithFormat:@"Achieved: %@", _releaseItem.achievement];
        else
            cell.achievementLabel.text = @"";
        
        for (UIGestureRecognizer* recognizer in cell.screenshotImageView.gestureRecognizers)
            [cell.screenshotImageView removeGestureRecognizer:recognizer];
        
        if (_releaseItem.screenshot != nil && _releaseItem.screenshot.length > 0)
        {
            [cell.screenshotImageView setImageWithURL:[NSURL URLWithString:_releaseItem.screenshot] placeholderImage:nil];
            [cell.screenshotImageView addGestureRecognizer:[[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(zoomImage:)]];
        }
        else
            cell.screenshotImageView.image = nil;
        
        UILongPressGestureRecognizer* longPress = [[UILongPressGestureRecognizer alloc] initWithTarget:self action:@selector(longPress:)];
        [cell.playButton addGestureRecognizer:longPress];
        
        BOOL has_downloads = _releaseItem.downloadURLs.count > 0;
        if (has_downloads && _allowDownload)
        {
            if (_selectedDownloadIndex != NSNotFound)
                cell.playButton.hidden = NO;
            else
                cell.playButton.hidden = YES;
        }
        else
            cell.playButton.hidden = YES;
        
        return cell;
    }
    
    NSMutableArray* comments = [self commentsArrayForSection:INTERFACE_IS_PHONE ? indexPath.section-1 : indexPath.section];
    CSCommentItem* comment = [comments objectAtIndex:indexPath.row];
    
    if (comment == nil)
        return nil;
    
    CommentItemCell* cell = [_commentsTableView dequeueReusableCellWithIdentifier:@"CommentItemCell"];

    if (cell.commentImageViews.count > 0)
    {
        [cell.commentImageViews makeObjectsPerformSelector:@selector(removeFromSuperview)];
        [cell.commentImageViews removeAllObjects];
    }
    
    cell.contentView.backgroundColor = ((indexPath.row & 1) == 0) ? [UIColor blackColor] : [UIColor colorWithWhite:0.1f alpha:1.0f];
    cell.handleLabel.text = comment.user.name;
    cell.dateLabel.text = comment.date;
    cell.commentLabel.attributedText = comment.text;
    cell.commentLabel.linkColor = [UIColor colorWithWhite:0.7f alpha:1.0f];
    cell.commentLabel.highlightedLinkColor = [UIColor colorWithWhite:0.9f alpha:1.0f];
    cell.commentLabel.delegate = self;
    
    if (comment.user.scener != nil)
        cell.photoImageView.image = comment.user.scener.photo;
    else
        cell.photoImageView.image = nil;

    cell.commentImageViews = [NSMutableArray array];
    
    NSUInteger index = 0;
    for (NSNumber* imageOffset in comment.imageOffsets)
    {
        NSString* commentText = [comment.text string];
        NSString* textBeforeImage = [commentText substringToIndex:[imageOffset unsignedIntegerValue]];
        CGSize size = [textBeforeImage sizeWithAttributes:[NSDictionary dictionaryWithObjectsAndKeys:[UIFont systemFontOfSize:13.0f], NSFontAttributeName, nil]];
        //CGSize size = [textBeforeImage sizeWithFont:[UIFont systemFontOfSize:13.0f] constrainedToSize:CGSizeMake(width, 100000.0f)];
        //CGRect rect = [textBeforeImage boundingRectWithSize:CGSizeMake(width, 10000.0f) options:NSStringDrawingUsesLineFragmentOrigin attributes:[NSDictionary dictionaryWithObjectsAndKeys:[UIFont systemFontOfSize:13.0f], NSFontAttributeName, nil] context:NULL];
        
        CGRect imageFrame = CGRectMake(64.0f, 32.0f + size.height, 160.0f, 100.0f);
        UIImageView* imageView = [[UIImageView alloc] initWithFrame:imageFrame];
        [imageView setImageWithURL:[comment.imageUrls objectAtIndex:index] placeholderImage:nil];
        imageView.contentMode = UIViewContentModeScaleAspectFit;
        imageView.backgroundColor = [UIColor blackColor];
        imageView.userInteractionEnabled = YES;
        [imageView addGestureRecognizer:[[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(zoomImage:)]];
        
        [cell.contentView addSubview:imageView];
        [cell.commentImageViews addObject:imageView];
        
        index++;
    }
    
    return cell;
}


// ----------------------------------------------------------------------------
- (CGFloat)tableView:(UITableView *)tableView heightForRowAtIndexPath:(NSIndexPath *)indexPath
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE)
    {
        if (indexPath.section == 0)
            return 495.0f;
    }
        
    NSMutableArray* comments = [self commentsArrayForSection:INTERFACE_IS_PHONE ? indexPath.section-1 : indexPath.section];
    CSCommentItem* comment = [comments objectAtIndex:indexPath.row];

    if (comment == nil)
        return 0;

    CommentItemCell* cell = [_commentsTableView dequeueReusableCellWithIdentifier:@"CommentItemCell"];
    CGFloat width = cell.commentLabel.frame.size.width;
    
    CGRect rect = [[comment.text string] boundingRectWithSize:CGSizeMake(width, 10000.0f) options:NSStringDrawingUsesLineFragmentOrigin attributes:[NSDictionary dictionaryWithObjectsAndKeys:[UIFont systemFontOfSize:13.0f], NSFontAttributeName, nil] context:NULL];

    CGFloat height = ceil(rect.size.height) + 45.0f;
    
    if (height > 2009.0f)
        height = 2009.0f;
    
    return height;
}


// ----------------------------------------------------------------------------
- (BOOL)tableView:(UITableView *)tableView canEditRowAtIndexPath:(NSIndexPath *)indexPath
// ----------------------------------------------------------------------------
{
    return NO;
}


// ----------------------------------------------------------------------------
- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE)
    {
        if (section == 0)
            return @"Release Info";
        
        return sCommentSections[section-1];
    }
    else
    {
        return sCommentSections[section];
    }
}


// ----------------------------------------------------------------------------
- (UIView*) tableView:(UITableView *)tableView viewForHeaderInSection:(NSInteger)section
// ----------------------------------------------------------------------------
{
    NSString *sectionTitle = [self tableView:tableView titleForHeaderInSection:section];
    if (sectionTitle == nil)
        return nil;
    
    CGFloat width = tableView.frame.size.width;
    
    UIView* customView = [[UIView alloc] init];
    customView.backgroundColor = [UIColor lightGrayColor]; //[UIColor colorWithPatternImage:[UIImage imageNamed:@"section_header"]];
    
    CGFloat left_inset = 15.0f;
    UILabel* headerLabel = [[UILabel alloc] initWithFrame:CGRectMake(left_inset, -23.0f, width - left_inset, 46.0f)];
    headerLabel.backgroundColor = [UIColor clearColor];
    headerLabel.autoresizingMask = UIViewAutoresizingFlexibleHeight | UIViewAutoresizingFlexibleWidth;
    headerLabel.opaque = NO;
    headerLabel.textColor = [UIColor darkGrayColor];
    headerLabel.font = [UIFont boldSystemFontOfSize:14.0f];
    headerLabel.textAlignment = NSTextAlignmentLeft;
    headerLabel.text = sectionTitle;
    [customView addSubview:headerLabel];
    return customView;
}


#pragma mark OHAttributedLabel delegate

// ----------------------------------------------------------------------------
-(UIColor*) colorForLink:(NSTextCheckingResult*)linkInfo underlineStyle:(int32_t*)underlineStyle ofText:(NSString*)text
// ----------------------------------------------------------------------------
{
    NSURL *url = [linkInfo URL];

    if (linkInfo.resultType != NSTextCheckingTypeLink || [[url scheme] isEqual:@"tel"])
    {
        *underlineStyle = kCTUnderlineStyleNone;
        return [UIColor whiteColor];
    }
    
    *underlineStyle = kCTUnderlineStyleSingle;
    return [UIColor colorWithWhite:0.7f alpha:1.0f];
}


#pragma mark UIScrollViewDelegate

    
// ----------------------------------------------------------------------------
- (UIView *)viewForZoomingInScrollView:(UIScrollView *)scrollView
// ----------------------------------------------------------------------------
{
    if (scrollView == _imageScrollView)
        return _zoomedScreenshotView;
    
    return nil;
}
    
@end



@implementation CommentItemCell

@synthesize handleLabel = _handleLabel;
@synthesize dateLabel = _dateLabel;
@synthesize commentLabel = _commentLabel;
@synthesize photoImageView = _photoImageView;
@synthesize commentImageViews = _commentImageViews;


@end


@implementation MainInfoItemCell

@synthesize titleLabel = _titleLabel;
@synthesize subtitleLabel = _subtitleLabel;
@synthesize groupsLabel = _groupsLabel;
@synthesize ratingLabel = _ratingLabel;
@synthesize screenshotImageView = _screenshotImageView;
@synthesize ratingView = _ratingView;
@synthesize ratingOverlayView = _ratingOverlayView;
@synthesize playButton = _playButton;
@synthesize releasedAtLabel = _releasedAtLabel;
@synthesize achievementLabel = _achievementLabel;


@end


