//
//  ScenerViewController.m
//  C64SceneBrowser
//
//  Created by Andreas Varga on 11.05.12.
//  Copyright (c) 2012 __MyCompanyName__. All rights reserved.
//

#import "ScenerViewController.h"
#import "CSScenerItem.h"
#import "CSReleaseItem.h"
#import "CSDatabaseCache.h"
#import "CSGroupItem.h"
#import "CSEventItem.h"
#import "CSCreditItem.h"
#import "MembershipView.h"
#import "MasterViewController.h"
#import "ReleaseViewController.h"
#import "CreditsViewController.h"
#import "MembershipViewController.h"
#import "UIImageView+WebCache.h"
#import "SSLoadingView.h"
#import "AppDelegate.h"

#import <CoreLocation/CoreLocation.h>


@implementation ScenerViewController


@synthesize scenerItem = _scenerItem;
@synthesize selectedReleaseItem = _selectedReleaseItem;
@synthesize selectedScenerItem = _selectedScenerItem;
@synthesize isAppearing = _isAppearing;


// ----------------------------------------------------------------------------
- (id)initWithNibName:(NSString *)nibNameOrNil bundle:(NSBundle *)nibBundleOrNil
// ----------------------------------------------------------------------------
{
    self = [super initWithNibName:nibNameOrNil bundle:nibBundleOrNil];
    if (self)
    {
        _previousScenerItem = nil;
        _actionSheet = nil;
        _selectedReleaseItem = nil;
        _membershipPopOver = nil;
        _scenerViewMode = SCENER_VIEW_CREDITS;
    }
    return self;
}


// ----------------------------------------------------------------------------
- (void)viewDidLoad
// ----------------------------------------------------------------------------
{
    [super viewDidLoad];

    _actionSheet = nil;
    _membershipPopOver = nil;
    _isAppearing = NO;
    
    [self updateView];
    
    if (_scenerItem != _previousScenerItem)
    {
        NSTimeInterval cache_age = 0.0f;
        NSTimeInterval timestamp = [[CSDatabaseCache sharedInstance] readTimeStamp:_scenerItem.identifier ofType:_scenerItem.type];
        NSTimeInterval currentTime = [NSDate timeIntervalSinceReferenceDate];
        cache_age = currentTime - timestamp;
        //NSLog(@"Cached scener item is %g seconds old", cache_age);
        
        if (_scenerItem.cached_depth < 2 || cache_age > 3600.0)
        {
            _releasesLoadingView = [[SSLoadingView alloc] initWithFrame:_scenerItemsTableView.bounds];
            [self.view addSubview:_releasesLoadingView];
            _scenerItemsTableView.hidden = YES;
            
            _membershipButton.enabled = NO;
            
            [_scenerItem cacheAndNotify:self withDepth:2];
        }
        
        [[CSDatabaseCache sharedInstance] writeTimeStampFile:_scenerItem];
        
        _previousScenerItem = _scenerItem;
    }
}


// ----------------------------------------------------------------------------
- (void)viewDidUnload
// ----------------------------------------------------------------------------
{
    [super viewDidUnload];
    // Release any retained subviews of the main view.
}


// ----------------------------------------------------------------------------
- (void) viewWillAppear:(BOOL)animated
// ----------------------------------------------------------------------------
{
    _isAppearing = YES;
    
    [super viewWillAppear:animated];
    
    UIViewController* masterViewController = [[self.splitViewController.viewControllers objectAtIndex:0] topViewController];
    
    if ([masterViewController class] == [CreditsViewController class])
    {
        CreditsViewController* creditsViewController = (CreditsViewController*)masterViewController;
        creditsViewController.selectedItem = _scenerItem;
    }
    
    AppDelegate* appDelegate = (AppDelegate*)[[UIApplication sharedApplication] delegate];
    if (appDelegate.masterPopoverController != nil)
        [self addMasterPopoverButton:appDelegate.masterPopoverBarButtonItem];
    
    _selectedReleaseItem = nil;
}


// ----------------------------------------------------------------------------
- (void) viewDidAppear:(BOOL)animated
// ----------------------------------------------------------------------------
{
    _isAppearing = NO;
}


// ----------------------------------------------------------------------------
- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
// ----------------------------------------------------------------------------
{
    return YES;
}


// ----------------------------------------------------------------------------
- (NSUInteger) supportedInterfaceOrientations
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE)
        return UIInterfaceOrientationMaskPortrait;
    else
        return UIInterfaceOrientationMaskAll;
}


// ----------------------------------------------------------------------------
- (UIInterfaceOrientation)preferredInterfaceOrientationForPresentation
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE)
        return UIInterfaceOrientationMaskPortrait;
    else
        return UIInterfaceOrientationMaskAll;
}


// ----------------------------------------------------------------------------
- (void) updateView
// ----------------------------------------------------------------------------
{
    [self updateViewModeButtons];
    
    NSString* handle_string = _scenerItem.handle;
    
    if (_scenerItem.alternativeHandle != nil)
        handle_string = [handle_string stringByAppendingFormat:@" (%@)", _scenerItem.alternativeHandle];

    _handleLabel.text = handle_string;
    //_handleLabel.text = _scenerItem.identifier;
    self.navigationItem.title = _scenerItem.handle;
    
    if (_scenerItem.photo != nil)
        [_photoView setImage:_scenerItem.photo];
    
    if (_scenerItem.country != nil)
    {
        CLGeocoder* geoCoder = [[CLGeocoder alloc] init];
        [geoCoder geocodeAddressString:_scenerItem.country completionHandler:^(NSArray *placemarks, NSError *error)
         {
             if (error == nil)
             {
                 CLPlacemark* placemark = [placemarks objectAtIndex:0];
                 MKCoordinateRegion region = MKCoordinateRegionMakeWithDistance(placemark.region.center, placemark.region.radius * 1.3f, placemark.region.radius * 1.3f);
                 [_mapView setRegion:region animated:NO];
                 _mapView.hidden = NO;
             }
             else
                 _mapView.hidden = YES;
         }];
    }
    else
        _mapView.hidden = YES;

    // Gather all non-ex groups
    NSString* groups_string = @"";
    BOOL first_group = YES;
    for (CSMemberOfItem* member in _scenerItem.memberOf)
    {
        if ([member.status isEqualToString:@"ex"])
            continue;
        
        if (first_group)
        {
            groups_string = [groups_string stringByAppendingString:member.group.name];
            first_group = NO;
        }
        else
            groups_string = [groups_string stringByAppendingFormat:@", %@", member.group.name];
    }
    
    _groupsLabel.text = groups_string;
 
    // Gather all unique professions
    NSMutableSet* professions = [NSMutableSet set];
    for (CSMemberOfItem* member in _scenerItem.memberOf)
        [professions addObjectsFromArray:member.professions];
    
    NSString* professions_string = @"";
    BOOL first_profession = YES;
    for (NSString* profession in professions)
    {
        if (first_profession)
        {
            professions_string = [professions_string stringByAppendingString:profession];
            first_profession = NO;
        }
        else
            professions_string = [professions_string stringByAppendingFormat:@", %@", profession];
    }

    _professionsLabel.text = professions_string;
    
    _membershipButton.enabled = YES;

    [_scenerItemsTableView reloadData];
}


// ----------------------------------------------------------------------------
- (void) updateViewModeButtons
// ----------------------------------------------------------------------------
{
    switch (_scenerViewMode)
    {
        case SCENER_VIEW_CREDITS:
            _creditsButton.backgroundColor = [UIColor colorWithWhite:0.1f alpha:1.0f];
            _releasesButton.backgroundColor = [UIColor clearColor];
            _eventsButton.backgroundColor = [UIColor clearColor];
            break;
            
        case SCENER_VIEW_RELEASES:
            _creditsButton.backgroundColor = [UIColor clearColor];
            _releasesButton.backgroundColor = [UIColor colorWithWhite:0.1f alpha:1.0f];
            _eventsButton.backgroundColor = [UIColor clearColor];
            break;
            
        case SCENER_VIEW_EVENTS:
            _creditsButton.backgroundColor = [UIColor clearColor];
            _releasesButton.backgroundColor = [UIColor clearColor];
            _eventsButton.backgroundColor = [UIColor colorWithWhite:0.1f alpha:1.0f];
            break;
    }
}


#pragma mark - CSDatabaseItemDelegate protocol


// ----------------------------------------------------------------------------
- (void) databaseItemIsCached:(CSDatabaseItem*)item
// ----------------------------------------------------------------------------
{
    //NSLog(@"Depth 2 finished loading! Reload comments and credits table!\n");
    
    [self updateView];
    
    [_releasesLoadingView removeFromSuperview];
    _scenerItemsTableView.hidden = NO;
}


// ----------------------------------------------------------------------------
- (void) cachingFailed
// ----------------------------------------------------------------------------
{
    [_releasesLoadingView removeFromSuperview];
    _scenerItemsTableView.hidden = NO;
    
    [[CSDatabaseCache sharedInstance] showNetworkError];
}


#pragma mark -

// ----------------------------------------------------------------------------
- (IBAction) tapActionButton:(id)sender
// ----------------------------------------------------------------------------
{
    if (_actionSheet != nil)
    {
        [_actionSheet dismissWithClickedButtonIndex:_actionSheet.cancelButtonIndex animated:YES];
        _actionSheet = nil;
        return;
    }
    
    if (_membershipPopOver != nil && _membershipPopOver.popoverVisible)
    {
        [_membershipPopOver dismissPopoverAnimated:YES];
        _membershipPopOver = nil;
    }
    
    _actionSheet = [[UIActionSheet alloc] initWithTitle:nil
                                               delegate:self
                                      cancelButtonTitle:nil
                                 destructiveButtonTitle:nil
                                      otherButtonTitles:@"Open in Safari", nil];
    
    _actionSheet.actionSheetStyle = UIActionSheetStyleBlackOpaque;
    [_actionSheet showFromBarButtonItem:sender animated:YES];
}


// ----------------------------------------------------------------------------
- (void) actionSheet:(UIActionSheet *)actionSheet didDismissWithButtonIndex:(NSInteger)buttonIndex
// ----------------------------------------------------------------------------
{
    //NSLog(@"Action sheet dismissed! button: %d\n", buttonIndex);
    switch (buttonIndex) 
    {
        case 0:
            [self openInSafari:self];
        default:
            break;
    }
    _actionSheet = nil;
}


// ----------------------------------------------------------------------------
- (IBAction) openInSafari:(id)sender
// ----------------------------------------------------------------------------
{
    NSString* scener_url_string = [NSString stringWithFormat:@"http://csdb.dk/scener/?id=%@", _scenerItem.identifier];
    [[UIApplication sharedApplication] openURL:[NSURL URLWithString:scener_url_string]];
}



// ----------------------------------------------------------------------------
- (IBAction) setScenerViewMode:(id)sender
// ----------------------------------------------------------------------------
{
    _scenerViewMode = (ScenerViewMode)[sender tag];

    [self updateViewModeButtons];
    [_scenerItemsTableView reloadData];
}



// ----------------------------------------------------------------------------
- (IBAction) clickMembershipButton:(id)sender
// ----------------------------------------------------------------------------
{
    if (_actionSheet != nil)
    {
        [_actionSheet dismissWithClickedButtonIndex:_actionSheet.cancelButtonIndex animated:YES];
        _actionSheet = nil;
    }
    
    if (_membershipPopOver != nil && _membershipPopOver.popoverVisible)
    {
        [_membershipPopOver dismissPopoverAnimated:YES];
        _membershipPopOver = nil;
    }
    else
        [self performSegueWithIdentifier:@"MembershipPopOver" sender:sender];
}


// ----------------------------------------------------------------------------
- (void)prepareForSegue:(UIStoryboardSegue *)segue sender:(id)sender
// ----------------------------------------------------------------------------
{
    if ([[segue identifier] isEqualToString:@"ShowReleaseDetail"])
    {
        //NSLog(@"%@", NSStringFromClass([[segue destinationViewController] class]));
        
        ReleaseViewController* releaseViewController = [segue destinationViewController];
        releaseViewController.releaseItem = _selectedReleaseItem;
    }
    else if ([[segue identifier] isEqualToString:@"ShowScenerDetail"])
    {
        ScenerViewController* scenerViewController = [segue destinationViewController];
        scenerViewController.scenerItem = _selectedScenerItem;
    }
    else if ([segue.identifier isEqualToString:@"MembershipPopOver"])
    {
        if (INTERFACE_IS_PHONE)
        {
            _membershipPopOver = nil;
            MembershipViewController* membershipViewController = [segue destinationViewController];
            [membershipViewController setScener:_scenerItem];
        }
        else
        {
            _membershipPopOver = ((UIStoryboardPopoverSegue*)segue).popoverController;
            MembershipViewController* membershipViewController = [segue destinationViewController];
            [membershipViewController setScener:_scenerItem];
            CGSize size = _membershipPopOver.popoverContentSize;
            size.width = MIN(membershipViewController.currentWidth, 900.0f);
            size.height = membershipViewController.currentHeight;
            _membershipPopOver.popoverContentSize = size;
        }
    }
}

// ----------------------------------------------------------------------------
- (void) addMasterPopoverButton:(UIBarButtonItem*)barButtonItem
// ----------------------------------------------------------------------------
{
    if (self.toolbarItems.count > 0 && [self.toolbarItems objectAtIndex:0] == barButtonItem)
        return;
    
    barButtonItem.title = @"Credits";
    NSMutableArray* toolbarItems = [NSMutableArray arrayWithObject:barButtonItem];
    [self setToolbarItems:[toolbarItems arrayByAddingObjectsFromArray:self.toolbarItems] animated:YES];
}


// ----------------------------------------------------------------------------
- (void) removeMasterPopoverButton:(UIBarButtonItem*)barButtonItem
// ----------------------------------------------------------------------------
{
    if (self.toolbarItems.count > 0 && [self.toolbarItems objectAtIndex:0] == barButtonItem)
        [self setToolbarItems:[self.toolbarItems subarrayWithRange:NSMakeRange(1, self.toolbarItems.count-1)] animated:YES];
}


#pragma mark TableView delegate and datasource

// ----------------------------------------------------------------------------
- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
// ----------------------------------------------------------------------------
{
    NSInteger count = 1;
    if (_scenerViewMode == SCENER_VIEW_EVENTS && _scenerItem.attendedEvents.count > 0 && _scenerItem.organizedEvents.count > 0)
        count = 2;

    if (INTERFACE_IS_PHONE)
        count++;
    
    return count;
}


// ----------------------------------------------------------------------------
- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE && section == 0)
        return 1;
    
    switch (_scenerViewMode)
    {
        case SCENER_VIEW_CREDITS:
            return _scenerItem.creditsDictionary.count;
            
        case SCENER_VIEW_RELEASES:
            return _scenerItem.released.count;
            
        case SCENER_VIEW_EVENTS:
            if (section == (INTERFACE_IS_PHONE ? 1 : 0))
                return _scenerItem.attendedEvents.count;
            else
                return _scenerItem.organizedEvents.count;
    }    

    return 0;
}


// ----------------------------------------------------------------------------
- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE && indexPath.section == 0)
    {
        ScenerInfoItemCell* cell = [_scenerItemsTableView dequeueReusableCellWithIdentifier:@"ScenerInfoItemCell"];
        
        [cell updateViewModeButtons:_scenerViewMode];
        
        NSString* handle_string = _scenerItem.handle;
        
        if (_scenerItem.alternativeHandle != nil)
            handle_string = [handle_string stringByAppendingFormat:@" (%@)", _scenerItem.alternativeHandle];
        
        cell.handleLabel.text = handle_string;
        
        if (_scenerItem.photo != nil)
            [cell.photoView setImage:_scenerItem.photo];
        
        // Gather all non-ex groups
        NSString* groups_string = @"";
        BOOL first_group = YES;
        for (CSMemberOfItem* member in _scenerItem.memberOf)
        {
            if ([member.status isEqualToString:@"ex"])
                continue;
            
            if (first_group)
            {
                groups_string = [groups_string stringByAppendingString:member.group.name];
                first_group = NO;
            }
            else
                groups_string = [groups_string stringByAppendingFormat:@", %@", member.group.name];
        }
        
        cell.groupsLabel.text = groups_string;
        
        // Gather all unique professions
        NSMutableSet* professions = [NSMutableSet set];
        for (CSMemberOfItem* member in _scenerItem.memberOf)
            [professions addObjectsFromArray:member.professions];
        
        NSString* professions_string = @"";
        BOOL first_profession = YES;
        for (NSString* profession in professions)
        {
            if (first_profession)
            {
                professions_string = [professions_string stringByAppendingString:profession];
                first_profession = NO;
            }
            else
                professions_string = [professions_string stringByAppendingFormat:@", %@", profession];
        }
        
        cell.professionsLabel.text = professions_string;
        
        return cell;
    }
    
    switch (_scenerViewMode)
    {
        case SCENER_VIEW_CREDITS:
        {
            ScenerCreditItemCell* cell = [tableView dequeueReusableCellWithIdentifier:@"ScenerCreditItemCell"];
            
            CSReleaseItem* releaseItem = [_scenerItem.sortedCreditReleases objectAtIndex:indexPath.row];
            
            cell.contentView.backgroundColor = ((indexPath.row & 1) == 0) ? [UIColor blackColor] : [UIColor colorWithWhite:0.1f alpha:1.0f];

            cell.nameLabel.text = releaseItem.name;
            cell.typeLabel.text = [NSString stringWithFormat:@"%ld %@", (long)releaseItem.releaseDate.year, releaseItem.releaseType];
            
            if (releaseItem.rating == nil || [releaseItem.rating isEqualToString:@""])
            {
                cell.ratingLabel.text = @"";
                cell.ratingStarsLabel.hidden = YES;
                cell.ratingStarsBackgroundLabel.hidden = YES;
            }
            else
            {
                cell.ratingLabel.text = [NSString stringWithFormat:@"%.1f/10", [releaseItem.rating floatValue]];
                
                float unscaled_width = cell.ratingStarsBackgroundLabel.frame.size.width;
                
                CGRect frame = cell.ratingStarsLabel.frame;
                CGRect frame2 = cell.ratingStarsBackgroundLabel.frame;
                
                frame.size.width = unscaled_width * ([releaseItem.rating floatValue] / 10.0f);
                frame.origin.x = cell.ratingLabel.frame.origin.x - unscaled_width - 10.0f;
                
                cell.ratingStarsLabel.frame = frame;
                cell.ratingStarsLabel.hidden = NO;
                
                frame2.origin.x = cell.ratingLabel.frame.origin.x - unscaled_width - 10.0f;
                
                cell.ratingStarsBackgroundLabel.frame = frame2;
                cell.ratingStarsBackgroundLabel.hidden = NO;
            }
            
            if (releaseItem.screenshot != nil && releaseItem.screenshot.length > 0)
            {
                NSURL* url = [NSURL URLWithString:releaseItem.screenshot];
                [cell.screenshotImageView setImageWithURL:url placeholderImage:nil];
            }
            else
                cell.screenshotImageView.image = nil;
            
            NSMutableArray* creditsOfRelease = [_scenerItem.creditsDictionary objectForKey:releaseItem.identifier];
            NSString* creditsString = @"";
            BOOL first_credit = YES;
            for (CSCreditItem* creditItem in creditsOfRelease)
            {
                if (first_credit)
                {
                    creditsString = [creditsString stringByAppendingString:creditItem.type];
                    first_credit = NO;
                }
                else
                    creditsString = [creditsString stringByAppendingFormat:@", %@", creditItem.type];
            }
            
            cell.creditsLabel.text = creditsString;
            
            return cell;
        }
            
        case SCENER_VIEW_RELEASES:
        {
            ReleaseItemCell* cell = [tableView dequeueReusableCellWithIdentifier:@"ReleaseItemCell"];
            
            cell.contentView.backgroundColor = ((indexPath.row & 1) == 0) ? [UIColor blackColor] : [UIColor colorWithWhite:0.1f alpha:1.0f];

            CSReleaseItem* releaseItem = [_scenerItem.released objectAtIndex:indexPath.row];
            if (releaseItem != nil)
            {
                cell.nameLabel.text = releaseItem.name;
                cell.typeLabel.text = [NSString stringWithFormat:@"%ld %@", (long)releaseItem.releaseDate.year, releaseItem.releaseType];
                
                if (releaseItem.screenshot != nil && releaseItem.screenshot.length > 0)
                {
                    NSURL* url = [NSURL URLWithString:releaseItem.screenshot];
                    [cell.screenshotImageView setImageWithURL:url placeholderImage:nil];
                }
                else
                    cell.screenshotImageView.image = nil;
                
                CSEventItem* event = releaseItem.releasedAtEvent;
                if (event != nil)
                    cell.eventLabel.text = [NSString stringWithFormat:@"Released at: %@", event.name];
                else
                    cell.eventLabel.text = @"";
                
                if (releaseItem.rating == nil || [releaseItem.rating isEqualToString:@""])
                {
                    cell.ratingLabel.text = @"";
                    cell.ratingStarsLabel.hidden = YES;
                    cell.ratingStarsBackgroundLabel.hidden = YES;
                }
                else
                {
                    cell.ratingLabel.text = [NSString stringWithFormat:@"%.1f/10", [releaseItem.rating floatValue]];
                    
                    float unscaled_width = cell.ratingStarsBackgroundLabel.frame.size.width;
                    
                    CGRect frame = cell.ratingStarsLabel.frame;
                    CGRect frame2 = cell.ratingStarsBackgroundLabel.frame;
                    
                    frame.size.width = unscaled_width * ([releaseItem.rating floatValue] / 10.0f);
                    frame.origin.x = cell.ratingLabel.frame.origin.x - unscaled_width - 10.0f;
                    
                    cell.ratingStarsLabel.frame = frame;
                    cell.ratingStarsLabel.hidden = NO;
                    
                    frame2.origin.x = cell.ratingLabel.frame.origin.x - unscaled_width - 10.0f;
                    
                    cell.ratingStarsBackgroundLabel.frame = frame2;
                    cell.ratingStarsBackgroundLabel.hidden = NO;
                }
                                                                                                                 }
            
            return cell;
        }
            
        case SCENER_VIEW_EVENTS:
        {
            EventItemCell* cell = [tableView dequeueReusableCellWithIdentifier:@"EventItemCell"];
            CSEventItem* eventItem = nil;
            
            cell.contentView.backgroundColor = ((indexPath.row & 1) == 0) ? [UIColor blackColor] : [UIColor colorWithWhite:0.1f alpha:1.0f];
            
            if (indexPath.section == (INTERFACE_IS_PHONE ? 1 : 0))
                eventItem = [_scenerItem.attendedEvents objectAtIndex:indexPath.row];
            else
                eventItem = [_scenerItem.organizedEvents objectAtIndex:indexPath.row];
            
            if (eventItem != nil)
            {
                cell.nameLabel.text = eventItem.name;
                cell.taglineLabel.text = eventItem.tagline;
            }
            
//            cell.datesLabel.text = [NSString stringWithFormat:@"%@ - %@", [[CSDatabaseCache sharedInstance] stringForDate:eventItem.startDate short:YES], [[CSDatabaseCache sharedInstance] stringForDate:eventItem.endDate short:YES]];
            
            return cell;
        }
    }
    
    return nil;
}


// ----------------------------------------------------------------------------
- (BOOL)tableView:(UITableView *)tableView canEditRowAtIndexPath:(NSIndexPath *)indexPath
// ----------------------------------------------------------------------------
{
    return NO;
}


// ----------------------------------------------------------------------------
- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE && indexPath.section == 0)
        return;
    
    CSReleaseItem* release = nil;
    
    switch (_scenerViewMode)
    {
        case SCENER_VIEW_CREDITS:
            release = [_scenerItem.sortedCreditReleases objectAtIndex:indexPath.row];
            break;
            
        case SCENER_VIEW_RELEASES:
            release = [_scenerItem.released objectAtIndex:indexPath.row];
            break;
            
        default:
            return;
    }

    if (release != nil && release != _selectedReleaseItem)
    {
        _selectedReleaseItem = release;
        [self performSegueWithIdentifier:@"ShowReleaseDetail" sender:self];
    }
}


// ----------------------------------------------------------------------------
- (CGFloat)tableView:(UITableView *)tableView heightForRowAtIndexPath:(NSIndexPath *)indexPath
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE && indexPath.section == 0)
        return 184.0f;
    
    switch (_scenerViewMode)
    {
        case SCENER_VIEW_CREDITS:
            return INTERFACE_IS_PHONE ? 78.0f : 52.0f;
            
        case SCENER_VIEW_RELEASES:
            return INTERFACE_IS_PHONE ? 78.0f : 52.0f;
            
        case SCENER_VIEW_EVENTS:
            return 52.0f;
    }
    
    return nil;
}


// ----------------------------------------------------------------------------
- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section
// ----------------------------------------------------------------------------
{
    if (INTERFACE_IS_PHONE)
    {
        if (section == 0)
            return @"Scener Info";
    }
    
    switch (_scenerViewMode)
    {
        case SCENER_VIEW_CREDITS:
            return @"Credits";            
            
        case SCENER_VIEW_RELEASES:
            return @"Releases released";            
            
        case SCENER_VIEW_EVENTS:
            if (section == (INTERFACE_IS_PHONE ? 1 : 0))
                return @"Events Attended";
            else
                return @"Events Organized";            
            
    }
    
    return nil;
}


// ----------------------------------------------------------------------------
- (UIView*) tableView:(UITableView *)tableView viewForHeaderInSection:(NSInteger)section
// ----------------------------------------------------------------------------
{
    NSString *sectionTitle = [self tableView:tableView titleForHeaderInSection:section];
    if (sectionTitle == nil)
        return nil;
    
    CGFloat width = tableView.frame.size.width;
    
    UIView* customView = [[UIView alloc] init];
    customView.backgroundColor = [UIColor lightGrayColor]; //[UIColor colorWithPatternImage:[UIImage imageNamed:@"section_header"]];
    
    CGFloat left_inset = 15.0f;
    UILabel* headerLabel = [[UILabel alloc] initWithFrame:CGRectMake(left_inset, -23.0f, width - left_inset, 46.0f)];
    headerLabel.backgroundColor = [UIColor clearColor];
    headerLabel.autoresizingMask = UIViewAutoresizingFlexibleHeight | UIViewAutoresizingFlexibleWidth;
    headerLabel.opaque = NO;
    headerLabel.textColor = [UIColor darkGrayColor];
    headerLabel.font = [UIFont boldSystemFontOfSize:14.0f];
    headerLabel.textAlignment = NSTextAlignmentLeft;
    headerLabel.text = sectionTitle;
    [customView addSubview:headerLabel];
    return customView;
}


@end



@implementation ReleaseItemCell

@synthesize nameLabel = _nameLabel;
@synthesize screenshotImageView = _screenshotImageView;
@synthesize typeLabel = _typeLabel;
@synthesize ratingLabel = _ratingLabel;
@synthesize ratingStarsLabel = _ratingStarsLabel;
@synthesize ratingStarsBackgroundLabel = _ratingStarsBackgroundLabel;

@end


@implementation ScenerCreditItemCell

@synthesize nameLabel = _nameLabel;
@synthesize screenshotImageView = _screenshotImageView;
@synthesize typeLabel = _typeLabel;
@synthesize creditsLabel = _creditsLabel;
@synthesize ratingLabel = _ratingLabel;
@synthesize ratingStarsLabel = _ratingStarsLabel;
@synthesize ratingStarsBackgroundLabel = _ratingStarsBackgroundLabel;

@end


@implementation EventItemCell

@synthesize nameLabel = _nameLabel;
@synthesize taglineLabel = _taglineLabel;
//@synthesize datesLabel = _datesLabel;

@end


@implementation ScenerInfoItemCell

@synthesize handleLabel = _handleLabel;
@synthesize groupsLabel = _groupsLabel;
@synthesize professionsLabel = _professionsLabel;
@synthesize photoView = _photoView;
@synthesize creditsButton = _creditsButton;
@synthesize releasesButton = _releasesButton;
@synthesize eventsButton = _eventsButton;

// ----------------------------------------------------------------------------
- (void) updateViewModeButtons:(ScenerViewMode)mode
// ----------------------------------------------------------------------------
{
    switch (mode)
    {
        case SCENER_VIEW_CREDITS:
            _creditsButton.backgroundColor = [UIColor colorWithWhite:0.1f alpha:1.0f];
            _releasesButton.backgroundColor = [UIColor clearColor];
            _eventsButton.backgroundColor = [UIColor clearColor];
            break;
            
        case SCENER_VIEW_RELEASES:
            _creditsButton.backgroundColor = [UIColor clearColor];
            _releasesButton.backgroundColor = [UIColor colorWithWhite:0.1f alpha:1.0f];
            _eventsButton.backgroundColor = [UIColor clearColor];
            break;
            
        case SCENER_VIEW_EVENTS:
            _creditsButton.backgroundColor = [UIColor clearColor];
            _releasesButton.backgroundColor = [UIColor clearColor];
            _eventsButton.backgroundColor = [UIColor colorWithWhite:0.1f alpha:1.0f];
            break;
    }
}


@end

